"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

STATE_NAME = "aws.s3.bucket_lifecycle"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    bucket: str,
    lifecycle_configuration: make_dataclass(
        "LifecycleConfiguration",
        [
            (
                "Rules",
                List[
                    make_dataclass(
                        "Rules",
                        [
                            ("ID", str),
                            ("Status", str),
                            (
                                "Filter",
                                make_dataclass(
                                    "Filter",
                                    [
                                        ("Prefix", str, field(default=None)),
                                        (
                                            "Tag",
                                            make_dataclass(
                                                "Tag",
                                                [
                                                    ("Key", str),
                                                    ("Value", str),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                        (
                                            "ObjectSizeGreaterThan",
                                            int,
                                            field(default=None),
                                        ),
                                        (
                                            "ObjectSizeLessThan",
                                            int,
                                            field(default=None),
                                        ),
                                        (
                                            "And",
                                            make_dataclass(
                                                "And",
                                                [
                                                    (
                                                        "Prefix",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "Tag",
                                                        make_dataclass(
                                                            "Tag",
                                                            [
                                                                ("Key", str),
                                                                ("Value", str),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                    ],
                                ),
                            ),
                            (
                                "Expiration",
                                make_dataclass(
                                    "Expiration",
                                    [
                                        ("Date", str, field(default=None)),
                                        ("Days", int, field(default=None)),
                                        (
                                            "ExpiredObjectDeleteMarker",
                                            bool,
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "Transition",
                                make_dataclass(
                                    "Transition",
                                    [
                                        ("Date", str, field(default=None)),
                                        ("Days", int, field(default=None)),
                                        ("StorageClass", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "NoncurrentVersionTransition",
                                make_dataclass(
                                    "NoncurrentVersionTransition",
                                    [
                                        ("NoncurrentDays", int, field(default=None)),
                                        ("StorageClass", str, field(default=None)),
                                        (
                                            "NewerNoncurrentVersions",
                                            int,
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "NoncurrentVersionExpiration",
                                make_dataclass(
                                    "NoncurrentVersionExpiration",
                                    [
                                        ("NoncurrentDays", int, field(default=None)),
                                        (
                                            "NewerNoncurrentVersions",
                                            int,
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "AbortIncompleteMultipartUpload",
                                make_dataclass(
                                    "AbortIncompleteMultipartUpload",
                                    [("DaysAfterInitiation", int, field(default=None))],
                                ),
                                field(default=None),
                            ),
                        ],
                    )
                ],
            )
        ],
    ) = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=4)),
                        ("max_attempts", int, field(default=30)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    Creates a lifecycle configuration for an S3 bucket resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The name of the S3 bucket in Amazon Web Services. It must be equal to the bucket parameter.
        bucket(string): The name of the S3 bucket in Amazon Web Services.
        lifecycle_configuration(Dict[str, Any], optional): The lifecycle configuration for the S3 bucket. Defaults to None.
            * Rules (List[Dict[str, Any]]): Specifies lifecycle configuration rules for an Amazon S3 bucket.
                * ID (str): Unique identifier for the rule. The value can't be longer than 255 characters.
                * Status (str): If Enabled, the rule is currently being applied. If Disabled, the rule is not currently being
                    applied.
                * Filter (Dict[str, Any]): The Filter is used to identify objects that a Lifecycle Rule applies to. A Filter must have exactly one of Prefix , Tag , or And specified.
                    * Prefix (str, optional): Prefix identifying one or more objects to which the rule applies.
                    * Tag (Dict[str, Any], optional): his tag must exist in the object's tag set in order for the rule to apply.
                        * Key (str): Name of the object key.
                        * Value (str): Value of the tag.
                    * ObjectSizeGreaterThan (int, optional): Minimum object size to which the rule applies.
                    * ObjectSizeLessThan (int, optional): Maximum object size to which the rule applies.
                    * And (Dict[str, Any], optional): This is used in a Lifecycle Rule Filter to apply a logical AND to two or more predicates. The Lifecycle Rule will apply to any object matching all of the predicates configured inside the And operator.
                        * Prefix (str, optional): Prefix identifying one or more objects to which the rule applies.
                        * Tag (Dict[str, Any], optional): his tag must exist in the object's tag set in order for the rule to apply.
                            * Key (str): Name of the object key.
                            * Value (str): Value of the tag.
                        * ObjectSizeGreaterThan (int, optional): Minimum object size to which the rule applies.
                        * ObjectSizeLessThan (int, optional): Maximum object size to which the rule applies.
                * Expiration (Dict[str, Any], optional): Specifies the expiration for the lifecycle of the object.
                    * Date (Text, optional): Indicates at what date the object is to be moved or deleted. Should be in GMT ISO 8601 Format.
                    * Days (int, optional): Indicates the lifetime, in days, of the objects that are subject to the rule. The value must be
                        a non-zero positive integer.
                    * ExpiredObjectDeleteMarker (bool, optional): Indicates whether Amazon S3 will remove a delete marker with no noncurrent versions. If set to
                        true, the delete marker will be expired; if set to false the policy takes no action. This cannot
                        be specified with Days or Date in a Lifecycle Expiration Policy.
                * Transition (Dict[str, Any], optional): Specifies when an object transitions to a specified storage class. For more information about
                    Amazon S3 lifecycle configuration rules, see Transitioning Objects Using Amazon S3 Lifecycle in
                    the Amazon S3 User Guide.
                    * Date (Text, optional): Indicates when objects are transitioned to the specified storage class. The date value must be
                        in ISO 8601 format. The time is always midnight UTC.
                    * Days (int, optional): Indicates the number of days after creation when objects are transitioned to the specified
                        storage class. The value must be a positive integer.
                    * StorageClass (str, optional): The storage class to which you want the object to transition.
                * NoncurrentVersionTransition (Dict[str, Any], optional): Container for the transition rule that describes when noncurrent objects transition to the
                    STANDARD_IA, ONEZONE_IA, INTELLIGENT_TIERING, GLACIER_IR, GLACIER, or DEEP_ARCHIVE storage
                    class. If your bucket is versioning-enabled (or versioning is suspended), you can set this
                    action to request that Amazon S3 transition noncurrent object versions to the STANDARD_IA,
                    ONEZONE_IA, INTELLIGENT_TIERING, GLACIER_IR, GLACIER, or DEEP_ARCHIVE storage class at a
                    specific period in the object's lifetime.
                    * NoncurrentDays (int, optional): Specifies the number of days an object is noncurrent before Amazon S3 can perform the associated
                        action. For information about the noncurrent days calculations, see How Amazon S3 Calculates How
                        Long an Object Has Been Noncurrent in the Amazon S3 User Guide.
                    * StorageClass (str, optional): The class of storage used to store the object.
                    * NewerNoncurrentVersions (int, optional): Specifies how many noncurrent versions Amazon S3 will retain. If there are this many more recent
                        noncurrent versions, Amazon S3 will take the associated action. For more information about
                        noncurrent versions, see Lifecycle configuration elements in the Amazon S3 User Guide.
                * NoncurrentVersionExpiration (Dict[str, Any], optional): Specifies when noncurrent object versions expire. Upon expiration, Amazon S3 permanently deletes
                    the noncurrent object versions. You set this lifecycle configuration action on a bucket that has
                    versioning enabled (or suspended) to request that Amazon S3 delete noncurrent object versions at
                    a specific period in the object's lifetime.
                    * NoncurrentDays (int, optional): Specifies the number of days an object is noncurrent before Amazon S3 can perform the associated
                        action. The value must be a non-zero positive integer. For information about the noncurrent days
                        calculations, see How Amazon S3 Calculates When an Object Became Noncurrent in the Amazon S3
                        User Guide.
                    * NewerNoncurrentVersions (int, optional): Specifies how many noncurrent versions Amazon S3 will retain. If there are this many more recent
                        noncurrent versions, Amazon S3 will take the associated action. For more information about
                        noncurrent versions, see Lifecycle configuration elements in the Amazon S3 User Guide.
                * AbortIncompleteMultipartUpload (Dict[str, Any], optional): Specifies the days since the initiation of an incomplete multipart upload that Amazon S3 will
                    wait before permanently removing all parts of the upload. For more information, see  Aborting
                    Incomplete Multipart Uploads Using a Bucket Lifecycle Policy in the Amazon S3 User Guide.
                    * DaysAfterInitiation (int, optional): Specifies the number of days after which Amazon S3 aborts an incomplete multipart upload.
        timeout(Dict, optional): Timeout configuration for S3 bucket lifecycle configuration.
            * update (string): Timeout configuration for updating the S3 bucket lifecycle configuration.
                * delay (int, optional): The amount of time in seconds to wait between attempts. Defaults to 4 seconds.
                * max_attempts (int, optional): Maximum attempts of waiting for the update. Defaults to 30 attempts.

    Request Syntax:
        [idem_test_aws_s3_bucket_lifecycle]:
          aws.s3.bucket_lifecycle.present:
            - name: 'string'
            - bucket: 'string'
            - lifecycle_configuration: {'string': []}

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_s3_bucket_lifecycle:
              aws.s3.bucket_lifecycle.present:
                - name: value
                - bucket: value
                - lifecycle_configuration:
                    Rules:
                    - ID: 'string'
                      Status: 'Enabled'|'Disabled'
                      Filter:
                        Prefix: ''
                      Expiration:
                        Date: datetime(2015, 1, 1)
                        Days: 123
                        ExpiredObjectDeleteMarker: True|False
                      Transition:
                        Date: datetime(2015, 1, 1)
                        Days: 123
                        StorageClass: 'GLACIER'|'STANDARD_IA'|'ONEZONE_IA'|'INTELLIGENT_TIERING'|'DEEP_ARCHIVE'|'GLACIER_IR'
                      NoncurrentVersionTransitions:
                        - NoncurrentDays: 123
                          StorageClass: 'GLACIER'|'STANDARD_IA'|'ONEZONE_IA'|'INTELLIGENT_TIERING'|'DEEP_ARCHIVE'|'GLACIER_IR'
                          NewerNoncurrentVersions: 123
                      NoncurrentVersionExpiration:
                        NoncurrentDays: 123
                        NewerNoncurrentVersions: 123
                      AbortIncompleteMultipartUpload:
                        DaysAfterInitiation: 123
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None

    if resource_id:
        if resource_id != bucket:
            result["result"] = False
            result["comment"] = (
                f"Bucket '{bucket}' and resource_id '{resource_id}' parameters must be the same",
            )
            return result

        before_ret = await hub.exec.aws.s3.bucket_lifecycle.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type=STATE_NAME, name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.s3.bucket_lifecycle.convert_raw_bucket_lifecycle_to_present(
            bucket=resource_id,
            raw_resource=before_ret.get("ret"),
            idem_resource_name=name,
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "lifecycle_configuration": lifecycle_configuration,
        }

        if hub.tool.aws.s3.bucket_lifecycle.is_bucket_lifecycle_configuration_identical(
            resource_parameters, result["old_state"]
        ):
            # no updates to apply
            return result

        if ctx.get("test", False):
            result["new_state"].update(resource_parameters)
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.would_update_comment(
                resource_type=STATE_NAME, name=name
            )
            return result
        else:
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "bucket": bucket,
                    "lifecycle_configuration": lifecycle_configuration,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=STATE_NAME, name=name
            )
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type=STATE_NAME, name=name
        )

    put_ret = await hub.exec.boto3.client.s3.put_bucket_lifecycle_configuration(
        ctx,
        Bucket=bucket,
        LifecycleConfiguration=lifecycle_configuration,
    )
    if not put_ret["result"]:
        result["result"] = False
        result["comment"] += put_ret["comment"]
        return result

    wait_for_updates_ret = await hub.tool.aws.s3.bucket_lifecycle.wait_for_updates(
        ctx,
        bucket=bucket,
        lifecycle_configuration=lifecycle_configuration,
        timeout=timeout.get("update") if timeout else None,
    )
    if not wait_for_updates_ret["result"]:
        result["result"] = False
        result["comment"] += wait_for_updates_ret["comment"]
        return result

    after_ret = await hub.exec.aws.s3.bucket_lifecycle.get(
        ctx, name=name, resource_id=bucket
    )
    result["new_state"] = copy.deepcopy(after_ret["ret"])

    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    r"""
    Deletes a lifecycle configuration from an S3 bucket resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The name of the S3 bucket in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.
        timeout(Dict, optional): Timeout configuration for S3 bucket lifecycle configuration.
            * delete (string): Timeout configuration for deleting the S3 bucket lifecycle configuration.
                * delay (int, optional): The amount of time in seconds to wait between attempts. Defaults to 4 seconds.
                * max_attempts (int, optional): Maximum attempts of waiting for the deletion. Defaults to 30 attempts.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_s3_bucket_lifecycle:
              aws.s3.bucket_lifecycle.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    before_ret = await hub.exec.aws.s3.bucket_lifecycle.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before_ret["result"]:
        if "NoSuchLifecycleConfiguration" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type=STATE_NAME, name=name
            )
        return result
    elif not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=STATE_NAME, name=name
        )
        return result
    result["old_state"] = before_ret["ret"]

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.s3.delete_bucket_lifecycle(
        ctx, Bucket=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    wait_for_delete_ret = await hub.tool.aws.s3.bucket_lifecycle.wait_for_delete(
        ctx,
        bucket=resource_id,
        timeout=timeout.get("delete") if timeout else None,
    )
    if not wait_for_delete_ret["result"]:
        result["result"] = False
        result["comment"] = wait_for_delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type=STATE_NAME, name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the lifecycle configuration for each S3 bucket under the given AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.bucket_lifecycle
    """

    result = {}

    list_buckets_ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not list_buckets_ret["result"]:
        hub.log.debug(f"Could not list S3 buckets: {list_buckets_ret['comment']}")
        return result

    for bucket in list_buckets_ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")

        get_bucket_lifecycle_ret = (
            await hub.exec.boto3.client.s3.get_bucket_lifecycle_configuration(
                ctx, Bucket=bucket_name
            )
        )
        if not get_bucket_lifecycle_ret["result"]:
            if "NoSuchLifecycleConfiguration" not in str(
                get_bucket_lifecycle_ret["comment"][0]
            ):
                hub.log.debug(
                    f"Could not get lifecycle configuration for S3 bucket '{bucket_name}': "
                    f"{get_bucket_lifecycle_ret['comment']}. Describe will skip this S3 bucket and continue."
                )
            continue

        resource_translated = (
            hub.tool.aws.s3.bucket_lifecycle.convert_raw_bucket_lifecycle_to_present(
                bucket=bucket_name,
                raw_resource=get_bucket_lifecycle_ret["ret"],
            )
        )

        result[resource_translated["name"]] = {
            f"{STATE_NAME}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



from typing import Any
from typing import Dict


def convert_raw_function_concurrency_config_to_present(
    hub,
    raw_resource: Dict[str, Any],
    function_name: str = None,
    idem_resource_name: str = None,
    function_arn: str = None,
) -> Dict[str, Any]:
    r"""
    Convert AWS returned data structure to correct idem function_provisioned_concurrency_config present state

     Args:
        hub: required for functions in hub
        raw_resource(Dict): The aws response to convert
        function_name (string, Optional): The name of the Lambda function
        idem_resource_name(string, optional): An Idem name of the resource.
        function_arn(string, optional): The Amazon Resource Name (ARN) of the function alias or version.

    Returns: Valid idem state for function_provisioned_concurrency_config of type Dict['string', Any]
    """

    function_arn_from_resource = raw_resource.get("FunctionArn")
    function_arn = (
        function_arn_from_resource if function_arn_from_resource else function_arn
    )

    resource_translated = {
        "name": idem_resource_name if idem_resource_name else function_arn,
        "resource_id": function_arn,
        "function_name": function_name,
    }

    provisioned_concurrent_executions = raw_resource.get(
        "RequestedProvisionedConcurrentExecutions"
    )
    if provisioned_concurrent_executions is not None:
        resource_translated[
            "provisioned_concurrent_executions"
        ] = provisioned_concurrent_executions

    qualifier = hub.tool.aws.arn_utils.get_qualifier(function_arn)
    if qualifier:
        resource_translated["qualifier"] = qualifier

    return resource_translated



"""State module for managing EC2 Snapshots"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    volume_id: str,
    resource_id: str = None,
    description: str = None,
    outpost_arn: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """Register an AWS Snapshot.

    Creates crash-consistent snapshots of multiple EBS volumes and stores the data in S3. Volumes are chosen by
    specifying an instance. Any attached volumes will produce one snapshot each that is crash-consistent across the
    instance. Boot volumes can be excluded by changing the parameters.  You can create multi-volume snapshots of
    instances in a Region and instances on an Outpost. If you create snapshots from an instance in a Region, the
    snapshots must be stored in the same Region as the instance. If you create snapshots from an instance on an
    Outpost, the snapshots can be stored on the same Outpost as the instance, or in the Region for that Outpost.

    Args:
        name (str):
            An Idem name for the snapshot resource.

        resource_id (str, optional):
            AWS Snapshot snapshot_id.

        volume_id (str, required):
            AWS Volume volume_id of the volume the snapshot is to be created from.

        description (str, optional):
            A description propagated to every snapshot specified by the instance. Defaults to None.

        outpost_arn (str, optional):
            The Amazon Resource Name (ARN) of the Outpost on which to create the local snapshots.   To
            create snapshots from an instance in a Region, omit this parameter. The snapshots are created in
            the same Region as the instance.   To create snapshots from an instance on an Outpost and store
            the snapshots in the Region, omit this parameter. The snapshots are created in the Region for
            the Outpost.   To create snapshots from an instance on an Outpost and store the snapshots on an
            Outpost, specify the ARN of the destination Outpost. The snapshots must be created on the same
            Outpost as the instance.   For more information, see  Create multi-volume local snapshots from
            instances on an Outpost in the Amazon Elastic Compute Cloud User Guide. Defaults to None.

        tags (dict, optional):
            Dict in the format of ``{tag-key: tag-value}`` to associate with the Snapshot.

            * Key (str):
                The key name that can be used to look up or retrieve the associated value.

            * Value (str):
                The value associated with this tag. For example tags with the key name of Name could have
                the AWS snapshot_id as the associated value.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws_auto.ec2.snapshot.present:
                - name: some-snapshot-ARN
                - resource_id: some-snapsshot-ARN
                - volume_id: some-volume-ARN
                - tags:
                    - Key: name
                      Value: ONE
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.snapshot.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = before["result"]
            return result

    if before:
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        old_tags = result["old_state"].get("tags")
        if tags is not None and tags != old_tags:
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result

        if update_ret["ret"]:
            result["comment"] = update_ret["comment"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if ctx.get("test", False) and update_ret["ret"] is not None:
            plan_state["tags"] = update_ret["ret"].get("tags")
            result["comment"] = (f"Would update tags for'{name}'",)
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "volume_id": volume_id,
                    "description": description,
                    "outpost_arn": outpost_arn,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.snapshot", name=name
            )
            return result

        ret = await hub.exec.boto3.client.ec2.create_snapshot(
            ctx,
            Description=description,
            OutpostArn=outpost_arn,
            VolumeId=volume_id,
            TagSpecifications=[
                {
                    "ResourceType": "snapshot",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["SnapshotId"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.snapshot", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            response = await hub.exec.aws.ec2.snapshot.get(
                ctx, name=name, resource_id=resource_id
            )
            after = response["ret"]
            result["new_state"] = copy.deepcopy(after)
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.ec2.snapshot", name=name
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    """Deregisters the specified Snapshot

    Deletes the specified snapshot. When you make periodic snapshots of a volume, the snapshots are incremental, and
    only the blocks on the device that have changed since your last snapshot are saved in the new snapshot. When you
    delete a snapshot, only the data not needed for any other snapshot is removed. So regardless of which prior
    snapshots have been deleted, all active snapshots will have access to all the information needed to restore the
    volume. You cannot delete a snapshot of the root device of an EBS volume used by a registered AMI. You must
    first de-register the AMI before you can delete the snapshot. For more information, see Delete an Amazon EBS
    snapshot in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name (str): An Idem name of the snapshot.
        resource_id(str): The snapshot ID.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws_auto.ec2.snapshot.absent:
                - name: value
                - resource_id: value
                - snapshot_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.snapshot", name=name
        )
        return result

    before = await hub.exec.aws.ec2.snapshot.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.snapshot", name=name
        )
    else:
        result["old_state"] = before["ret"]

        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.snapshot", name=name
            )
            return result

        ret = await hub.exec.boto3.client.ec2.delete_snapshot(
            ctx, SnapshotId=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.ec2.snapshot", name=name
            )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets information about the AWS Snapshot(s)

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws_auto.ec2.snapshot
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_snapshots(ctx, OwnerIds=["self"])

    if not ret["result"]:
        hub.log.debug(f"Could not describe snapshot {ret['comment']}")
        return {}

    for snapshot in ret["ret"]["Snapshots"]:
        resource_id = snapshot.get("SnapshotId")
        resource_translated = (
            await hub.tool.aws.ec2.conversion_utils.convert_raw_snapshot_to_present(
                ctx=ctx, raw_resource=snapshot, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.snapshot.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



from typing import Dict

"""
Exec functions for AWS API Gateway v2 Integration resources.
"""


async def get(hub, ctx, resource_id: str, api_id: str, name: str = None) -> Dict:
    """
    Get a single ApiGatewayV2 Authorizer resource from AWS.

    Args:
        resource_id(string): The AWS api gateway v2 integration id.
        api_id(string): The AWS api gateway v2 API id.
        name(string, optional): The name of the Idem state for logging.
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.apigatewayv2.get_integration(
        ctx, ApiId=api_id, IntegrationId=resource_id
    )
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.integration", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    result[
        "ret"
    ] = hub.tool.aws.apigatewayv2.integration.convert_raw_integration_to_present(
        api_id=api_id,
        raw_resource=ret["ret"],
        idem_resource_name=name if name is not None else resource_id,
    )
    return result



from collections import OrderedDict
from typing import Dict


async def update_addon(
    hub,
    ctx,
    name: str,
    before: Dict,
    addon_version: str,
    service_account_role_arn: str,
    resolve_conflicts: str,
    client_request_token: str,
    timeout: Dict = None,
):
    """
    Updates an Amazon EKS add-on.

    Args:
       hub:
       ctx:
       name: An Idem name of the resource
       before(Dict): AWS cluster addon
       addon_version(Text): The version of the add-on. The version must match one of the versions returned
                            by ` DescribeAddonVersions
       service_account_role_arn(Text): The Amazon Resource Name (ARN) of an existing IAM role to bind to the add-on's
                                       service account. The role must be assigned the IAM permissions required by the add-on. If you don't specify
                                       an existing IAM role, then the add-on uses the permissions assigned to the node IAM role.
       resolve_conflicts(Text): How to resolve parameter value conflicts when migrating an existing add-on
                                to an Amazon EKS add-on
       client_request_token(Text):Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
                                  This field is autopopulated if not provided.
       timeout(Dict, optional): Timeout configuration for creating or updating addon.
            * create (Dict) -- Timeout configuration for creating addon
                * delay(int, default=10) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=60) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating addon
                * delay(int, default=10) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=60) -- Customized timeout configuration containing delay and max attempts.

    Returns:
       {"result": True|False, "comment": A message Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}

    if (
        service_account_role_arn
        and before.get("service_account_role_arn") != service_account_role_arn
    ):
        update_payload["serviceAccountRoleArn"] = service_account_role_arn
    if addon_version and before.get("addon_version") != addon_version:
        update_payload["addonVersion"] = addon_version
    if resolve_conflicts and before.get("resolve_conflicts") != resolve_conflicts:
        update_payload["resolveConflicts"] = resolve_conflicts
    if client_request_token:
        update_payload["clientRequestToken"] = client_request_token

    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.eks.update_addon(
                ctx=ctx,
                clusterName=before.get("cluster_name"),
                addonName=before.get("resource_id"),
                **update_payload,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            else:
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=10,
                    default_max_attempts=60,
                    timeout_config=timeout.get("update") if timeout else None,
                )
                hub.log.debug(f"Waiting on updating aws.eks.addon '{name}'")
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "eks",
                        "addon_active",
                        addonName=before.get("resource_id"),
                        clusterName=before.get("cluster_name"),
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
                    return result
        result["ret"] = {}
        update_params_map = OrderedDict(
            {
                "addonVersion": "addon_version",
                "serviceAccountRoleArn": "service_account_role_arn",
                "resolveConflicts": "resolve_conflicts",
            }
        )

        for parameter_raw, parameter_present in update_params_map.items():
            if parameter_raw in update_payload:
                result["ret"][parameter_present] = update_payload.get(parameter_raw)
                result["comment"] = result["comment"] + (
                    f"Update {parameter_present}: {update_payload[parameter_raw]} on addon {before.get('resource_id')}",
                )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_load_balancer_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, str] = None,
    attributes: List[Dict[str, str]] = None,
) -> Dict[str, Any]:
    r"""
    Util function to convert raw resource state from AWS ElasticLoadBalancingv2 Load Balancer to present input format.

    Args:
        raw_resource(Dict[str, Any]):
            The AWS response from 'ElasticLoadBalancingv2 describe_load_balancers' to convert.
        idem_resource_name(Text, optional): An Idem name of the resource.
        tags(Dict[str, str], optional): The AWS ElasticLoadBalancingv2 Load Balancer tags.
        attributes(List[Dict[str, str]], optional):
            List of attributes associated with given ElasticLoadBalancingv2 Load Balancer.

    Returns:
        Dict[str, Any]
    """

    resource_id = raw_resource.get("LoadBalancerArn")
    resource_parameters = OrderedDict(
        {
            "SecurityGroups": "security_groups",
            "Scheme": "scheme",
            "Type": "lb_type",
            "IpAddressType": "ip_address_type",
            "CustomerOwnedIpv4Pool": "customer_owned_ipv4_pool",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    subnets = []
    subnet_mappings = []
    if raw_resource.get("AvailabilityZones"):
        for zone in raw_resource.get("AvailabilityZones"):
            subnets.append(zone.get("SubnetId"))
            if zone.get("LoadBalancerAddresses"):
                mapping_parameters = OrderedDict(
                    {
                        "SubnetId": zone.get("SubnetId"),
                        "AllocationId": zone.get("LoadBalancerAddresses")[0].get(
                            "AllocationId"
                        ),
                        "PrivateIPv4Address": zone.get("LoadBalancerAddresses")[0].get(
                            "PrivateIPv4Address"
                        ),
                        "IPv6Address": zone.get("LoadBalancerAddresses")[0].get(
                            "IPv6Address"
                        ),
                    }
                )
                mapping = {}
                for name, value in mapping_parameters.items():
                    if value:
                        mapping[name] = value
                if mapping:
                    subnet_mappings.append(mapping)
    resource_translated["subnets"] = subnets
    resource_translated["subnet_mappings"] = subnet_mappings

    if attributes:
        resource_translated["attributes"] = attributes
    if tags:
        resource_translated["tags"] = tags
    return resource_translated


def convert_raw_target_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, str] = None,
    attributes: List[Dict[str, str]] = None,
    targets: List[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    r"""
    Util function to convert raw resource state from ElasticLoadBalancingv2 Target Group to present input format.

    Args:
        raw_resource(Dict[str, Any]): The AWS response from 'ElasticLoadBalancingv2 describe_target_groups' to convert.
        idem_resource_name(Text, optional): An Idem name of the resource.
        tags(Dict[str, str], optional): The AWS ElasticLoadBalancingv2 Target Group tags.
        attributes(List[Dict[str, str]], optional):
            List of attributes associated with given ElasticLoadBalancingv2 Target Group.
        targets(List[Dict[str, Any]], optional):
            List of targets associated with given ElasticLoadBalancingv2 Target Group.

    Returns:
        Dict[str, Any]
    """
    resource_id = raw_resource.get("TargetGroupArn")
    resource_parameters = OrderedDict(
        {
            "Protocol": "protocol",
            "Port": "port",
            "VpcId": "vpc_id",
            "HealthCheckProtocol": "health_check_protocol",
            "HealthCheckPort": "health_check_port",
            "HealthCheckEnabled": "health_check_enabled",
            "HealthCheckIntervalSeconds": "health_check_interval_seconds",
            "HealthCheckTimeoutSeconds": "health_check_timeout_seconds",
            "HealthyThresholdCount": "healthy_threshold_count",
            "UnhealthyThresholdCount": "unhealthy_threshold_count",
            "HealthCheckPath": "health_check_path",
            "Matcher": "matcher",
            "TargetType": "target_type",
            "ProtocolVersion": "protocol_version",
            "IpAddressType": "ip_address_type",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if attributes:
        resource_translated["attributes"] = attributes
    if tags:
        resource_translated["tags"] = tags

    targets_add = []
    if targets:
        for target in targets:
            targets_add.append(target.get("Target"))
    resource_translated["targets"] = targets_add

    return resource_translated


def convert_raw_listener_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, str] = None,
    certificates: List[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    r"""
    Util functions to convert raw resource state from AWS ELBv2 Listener to present input format.

    Args:
        raw_resource(dict[str, Any]):
            The AWS response from 'ElasticLoadBalancingv2 describe_listeners' to convert.

        idem_resource_name(str, optional):
            An Idem name of the resource.

        tags(dict[str, str], optional):
            The AWS ElasticLoadBalancingv2 Listener tags.

        certificates(list[dict[str, Any]], optional):
            List of certificates associated with given ElasticLoadBalancingv2 Listener.

    Returns:
        Dict[str, Any]
    """
    resource_id = raw_resource.get("ListenerArn")
    resource_parameters = OrderedDict(
        {
            "Protocol": "protocol",
            "Port": "port",
            "SslPolicy": "ssl_policy",
            "Certificates": "certificates",
            "AlpnPolicy": "alpn_policy",
            "LoadBalancerArn": "lb_arn",
            "DefaultActions": "default_actions",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if tags:
        resource_translated["tags"] = tags

    default_certificates = certificate_list = []
    if certificates:
        for certificate in certificates:
            is_default = certificate.get("IsDefault", False)
            certificate.pop("IsDefault", None)
            if is_default:
                default_certificates.append(certificate)
            else:
                certificate_list.append(certificate)
        resource_translated["certificates"] = certificate_list
        resource_translated["default_certificates"] = default_certificates
    return resource_translated



"""
https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/spot-best-practices.html#which-spot-request-method-to-use

Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

Note..

The describe function uses the describe_spot_instance_requests to request spot instance details.
The present function uses the request_spot_instances to request creation of a spot instance.
The delete function uses the standard terminate_instances as the cancel_spot_requests does
not terminate an instance that comes online.

If you wish to delete or terminate an instance, and it's a spot instance, you must use the describe
function here to obtain the InstanceID and then use ec2.instance.absent to fully terminate the instance.
"""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource", "soft_fail"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    launch_specification: make_dataclass(
        "RequestSpotLaunchSpecification",
        [
            ("SecurityGroupIds", List[str], field(default=None)),
            ("SecurityGroups", List[str], field(default=None)),
            ("AddressingType", str, field(default=None)),
            (
                "BlockDeviceMappings",
                List[
                    make_dataclass(
                        "BlockDeviceMapping",
                        [
                            ("DeviceName", str, field(default=None)),
                            ("VirtualName", str, field(default=None)),
                            (
                                "Ebs",
                                make_dataclass(
                                    "EbsBlockDevice",
                                    [
                                        (
                                            "DeleteOnTermination",
                                            bool,
                                            field(default=None),
                                        ),
                                        ("Iops", int, field(default=None)),
                                        ("SnapshotId", str, field(default=None)),
                                        ("VolumeSize", int, field(default=None)),
                                        ("VolumeType", str, field(default=None)),
                                        ("KmsKeyId", str, field(default=None)),
                                        ("Throughput", int, field(default=None)),
                                        ("OutpostArn", str, field(default=None)),
                                        ("Encrypted", bool, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("NoDevice", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("EbsOptimized", bool, field(default=None)),
            (
                "IamInstanceProfile",
                make_dataclass(
                    "IamInstanceProfileSpecification",
                    [
                        ("Arn", str, field(default=None)),
                        ("Name", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("ImageId", str, field(default=None)),
            ("InstanceType", str, field(default=None)),
            ("KernelId", str, field(default=None)),
            ("KeyName", str, field(default=None)),
            (
                "Monitoring",
                make_dataclass("RunInstancesMonitoringEnabled", [("Enabled", bool)]),
                field(default=None),
            ),
            (
                "NetworkInterfaces",
                List[
                    make_dataclass(
                        "InstanceNetworkInterfaceSpecification",
                        [
                            ("AssociatePublicIpAddress", bool, field(default=None)),
                            ("DeleteOnTermination", bool, field(default=None)),
                            ("Description", str, field(default=None)),
                            ("DeviceIndex", int, field(default=None)),
                            ("Groups", List[str], field(default=None)),
                            ("Ipv6AddressCount", int, field(default=None)),
                            (
                                "Ipv6Addresses",
                                List[
                                    make_dataclass(
                                        "InstanceIpv6Address",
                                        [("Ipv6Address", str, field(default=None))],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("NetworkInterfaceId", str, field(default=None)),
                            ("PrivateIpAddress", str, field(default=None)),
                            (
                                "PrivateIpAddresses",
                                List[
                                    make_dataclass(
                                        "PrivateIpAddressSpecification",
                                        [
                                            ("Primary", bool, field(default=None)),
                                            (
                                                "PrivateIpAddress",
                                                str,
                                                field(default=None),
                                            ),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "SecondaryPrivateIpAddressCount",
                                int,
                                field(default=None),
                            ),
                            ("SubnetId", str, field(default=None)),
                            ("AssociateCarrierIpAddress", bool, field(default=None)),
                            ("InterfaceType", str, field(default=None)),
                            ("NetworkCardIndex", int, field(default=None)),
                            (
                                "Ipv4Prefixes",
                                List[
                                    make_dataclass(
                                        "Ipv4PrefixSpecificationRequest",
                                        [("Ipv4Prefix", str, field(default=None))],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("Ipv4PrefixCount", int, field(default=None)),
                            (
                                "Ipv6Prefixes",
                                List[
                                    make_dataclass(
                                        "Ipv6PrefixSpecificationRequest",
                                        [("Ipv6Prefix", str, field(default=None))],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("Ipv6PrefixCount", int, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            (
                "Placement",
                make_dataclass(
                    "SpotPlacement",
                    [
                        ("AvailabilityZone", str, field(default=None)),
                        ("GroupName", str, field(default=None)),
                        ("Tenancy", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("RamdiskId", str, field(default=None)),
            ("SubnetId", str, field(default=None)),
            ("UserData", str, field(default=None)),
        ],
    ) = None,
    launch_group: str = None,
    spot_price: str = None,
    type: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    # The following parameters can only be specified at creation
    availability_zone_group: str = None,
    instance_count: int = None,
    instance_interruption_behavior: str = None,
    valid_from: str = None,
    valid_until: str = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    **Autogenerated function**

    Args:
        hub:
        ctx:
        name(Text): An Idem name of the resource.
        resource_id(Text): The spot instance request id
        availability_zone_group(Text): The user-specified name for a logical grouping of requests.
            When you specify an Availability Zone group in a Spot Instance request,
            all Spot Instances in the request are launched in the same Availability Zone.
            Instance proximity is maintained with this parameter, but the choice of
            Availability Zone is not. The group applies only to requests for Spot Instances
            of the same instance type. Any additional Spot Instance requests that are
            specified with the same Availability Zone group name are launched in that
            same Availability Zone, as long as at least one instance from the group is still active.

            If there is no active instance running in the Availability Zone group that
            you specify for a new Spot Instance request (all instances are terminated,
            the request is expired, or the maximum price you specified falls below
            current Spot price), then Amazon EC2 launches the instance in any Availability
            Zone where the constraint can be met. Consequently, the subsequent set of Spot
            Instances could be placed in a different zone from the original request,
            even if you specified the same Availability Zone group.

            Default: Instances are launched in any available Availability Zone.
        instance_count(int): The maximum number of Spot Instances to launch.
        instance_interruption_behavior(str): The behavior when a Spot Instance is interrupted. The default is terminate.
            Valid Values: hibernate | stop | terminate
        launch_group(str):
            The instance launch group. Launch groups are Spot Instances that launch together and terminate together.
            Default: Instances are launched and terminated individually
        launch_specification(Dict[str, Any], optional): The launch specification. Defaults to None.
            * SecurityGroupIds (List[str], optional): One or more security group IDs.
            * SecurityGroups (List[str], optional): One or more security groups. When requesting instances in a VPC, you must specify the IDs of the
                security groups. When requesting instances in EC2-Classic, you can specify the names or the IDs
                of the security groups.
            * AddressingType (str, optional): Deprecated.
            * BlockDeviceMappings (List[Dict[str, Any]], optional): One or more block device mapping entries. You can't specify both a snapshot ID and an encryption
                value. This is because only blank volumes can be encrypted on creation. If a snapshot is the
                basis for a volume, it is not blank and its encryption status is used for the volume encryption
                status.
                * DeviceName (str, optional): The device name (for example, /dev/sdh or xvdh).
                * VirtualName (str, optional): The virtual device name (ephemeralN). Instance store volumes are numbered starting from 0. An
                    instance type with 2 available instance store volumes can specify mappings for ephemeral0 and
                    ephemeral1. The number of available instance store volumes depends on the instance type. After
                    you connect to the instance, you must mount the volume. NVMe instance store volumes are
                    automatically enumerated and assigned a device name. Including them in your block device mapping
                    has no effect. Constraints: For M3 instances, you must specify instance store volumes in the
                    block device mapping for the instance. When you launch an M3 instance, we ignore any instance
                    store volumes specified in the block device mapping for the AMI.
                * Ebs (Dict[str, Any], optional): Parameters used to automatically set up EBS volumes when the instance is launched.
                    * DeleteOnTermination (bool, optional): Indicates whether the EBS volume is deleted on instance termination. For more information, see
                        Preserving Amazon EBS volumes on instance termination in the Amazon EC2 User Guide.
                    * Iops (int, optional): The number of I/O operations per second (IOPS). For gp3, io1, and io2 volumes, this represents
                        the number of IOPS that are provisioned for the volume. For gp2 volumes, this represents the
                        baseline performance of the volume and the rate at which the volume accumulates I/O credits for
                        bursting. The following are the supported values for each volume type:    gp3: 3,000-16,000 IOPS
                        io1: 100-64,000 IOPS    io2: 100-64,000 IOPS   For io1 and io2 volumes, we guarantee 64,000 IOPS
                        only for Instances built on the Nitro System. Other instance families guarantee performance up
                        to 32,000 IOPS. This parameter is required for io1 and io2 volumes. The default for gp3 volumes
                        is 3,000 IOPS. This parameter is not supported for gp2, st1, sc1, or standard volumes.
                    * SnapshotId (str, optional): The ID of the snapshot.
                    * VolumeSize (int, optional): The size of the volume, in GiBs. You must specify either a snapshot ID or a volume size. If you
                        specify a snapshot, the default is the snapshot size. You can specify a volume size that is
                        equal to or larger than the snapshot size. The following are the supported volumes sizes for
                        each volume type:    gp2 and gp3:1-16,384    io1 and io2: 4-16,384    st1 and sc1: 125-16,384
                        standard: 1-1,024
                    * VolumeType (str, optional): The volume type. For more information, see Amazon EBS volume types in the Amazon EC2 User Guide.
                        If the volume type is io1 or io2, you must specify the IOPS that the volume supports.
                    * KmsKeyId (str, optional): Identifier (key ID, key alias, ID ARN, or alias ARN) for a customer managed CMK under which the
                        EBS volume is encrypted. This parameter is only supported on BlockDeviceMapping objects called
                        by RunInstances, RequestSpotFleet, and RequestSpotInstances.
                    * Throughput (int, optional): The throughput that the volume supports, in MiB/s. This parameter is valid only for gp3 volumes.
                        Valid Range: Minimum value of 125. Maximum value of 1000.
                    * OutpostArn (str, optional): The ARN of the Outpost on which the snapshot is stored. This parameter is only supported on
                        BlockDeviceMapping objects called by  CreateImage.
                    * Encrypted (bool, optional): Indicates whether the encryption state of an EBS volume is changed while being restored from a
                        backing snapshot. The effect of setting the encryption state to true depends on the volume
                        origin (new or from a snapshot), starting encryption state, ownership, and whether encryption by
                        default is enabled. For more information, see Amazon EBS encryption in the Amazon EC2 User
                        Guide. In no case can you remove encryption from an encrypted volume. Encrypted volumes can only
                        be attached to instances that support Amazon EBS encryption. For more information, see Supported
                        instance types. This parameter is not returned by DescribeImageAttribute.
                * NoDevice (str, optional): To omit the device from the block device mapping, specify an empty string. When this property is
                    specified, the device is removed from the block device mapping regardless of the assigned value.
            * EbsOptimized (bool, optional): Indicates whether the instance is optimized for EBS I/O. This optimization provides dedicated
                throughput to Amazon EBS and an optimized configuration stack to provide optimal EBS I/O
                performance. This optimization isn't available with all instance types. Additional usage charges
                apply when using an EBS Optimized instance. Default: false
            * IamInstanceProfile (Dict[str, Any], optional): The IAM instance profile.
                * Arn (str, optional): The Amazon Resource Name (ARN) of the instance profile.
                * Name (str, optional): The name of the instance profile.
            * ImageId (str, optional): The ID of the AMI.
            * InstanceType (str, optional): The instance type. Only one instance type can be specified.
            * KernelId (str, optional): The ID of the kernel.
            * KeyName (str, optional): The name of the key pair.
            * Monitoring (Dict[str, Any], optional): Indicates whether basic or detailed monitoring is enabled for the instance. Default: Disabled
                * Enabled (bool): Indicates whether detailed monitoring is enabled. Otherwise, basic monitoring is enabled.
            * NetworkInterfaces (List[Dict[str, Any]], optional): One or more network interfaces. If you specify a network interface, you must specify subnet IDs
                and security group IDs using the network interface.
                * AssociatePublicIpAddress (bool, optional): Indicates whether to assign a public IPv4 address to an instance you launch in a VPC. The public
                    IP address can only be assigned to a network interface for eth0, and can only be assigned to a
                    new network interface, not an existing one. You cannot specify more than one network interface
                    in the request. If launching into a default subnet, the default value is true.
                * DeleteOnTermination (bool, optional): If set to true, the interface is deleted when the instance is terminated. You can specify true
                    only if creating a new network interface when launching an instance.
                * Description (str, optional): The description of the network interface. Applies only if creating a network interface when
                    launching an instance.
                * DeviceIndex (int, optional): The position of the network interface in the attachment order. A primary network interface has a
                    device index of 0. If you specify a network interface when launching an instance, you must
                    specify the device index.
                * Groups (List[str], optional): The IDs of the security groups for the network interface. Applies only if creating a network
                    interface when launching an instance.
                * Ipv6AddressCount (int, optional): A number of IPv6 addresses to assign to the network interface. Amazon EC2 chooses the IPv6
                    addresses from the range of the subnet. You cannot specify this option and the option to assign
                    specific IPv6 addresses in the same request. You can specify this option if you've specified a
                    minimum number of instances to launch.
                * Ipv6Addresses (List[Dict[str, Any]], optional): One or more IPv6 addresses to assign to the network interface. You cannot specify this option
                    and the option to assign a number of IPv6 addresses in the same request. You cannot specify this
                    option if you've specified a minimum number of instances to launch.
                    * Ipv6Address (str, optional): The IPv6 address.
                * NetworkInterfaceId (str, optional): The ID of the network interface. If you are creating a Spot Fleet, omit this parameter because
                    you can’t specify a network interface ID in a launch specification.
                * PrivateIpAddress (str, optional): The private IPv4 address of the network interface. Applies only if creating a network interface
                    when launching an instance. You cannot specify this option if you're launching more than one
                    instance in a RunInstances request.
                * PrivateIpAddresses (List[Dict[str, Any]], optional): One or more private IPv4 addresses to assign to the network interface. Only one private IPv4
                    address can be designated as primary. You cannot specify this option if you're launching more
                    than one instance in a RunInstances request.
                    * Primary (bool, optional): Indicates whether the private IPv4 address is the primary private IPv4 address. Only one IPv4
                        address can be designated as primary.
                    * PrivateIpAddress (str, optional): The private IPv4 addresses.
                * SecondaryPrivateIpAddressCount (int, optional): The number of secondary private IPv4 addresses. You can't specify this option and specify more
                    than one private IP address using the private IP addresses option. You cannot specify this
                    option if you're launching more than one instance in a RunInstances request.
                * SubnetId (str, optional): The ID of the subnet associated with the network interface. Applies only if creating a network
                    interface when launching an instance.
                * AssociateCarrierIpAddress (bool, optional): Indicates whether to assign a carrier IP address to the network interface. You can only assign a
                    carrier IP address to a network interface that is in a subnet in a Wavelength Zone. For more
                    information about carrier IP addresses, see Carrier IP addresses in the Amazon Web Services
                    Wavelength Developer Guide.
                * InterfaceType (str, optional): The type of network interface. Valid values: interface | efa
                * NetworkCardIndex (int, optional): The index of the network card. Some instance types support multiple network cards. The primary
                    network interface must be assigned to network card index 0. The default is network card index 0.
                    If you are using RequestSpotInstances to create Spot Instances, omit this parameter because you
                    can’t specify the network card index when using this API. To specify the network card index, use
                    RunInstances.
                * Ipv4Prefixes (List[Dict[str, Any]], optional): One or more IPv4 delegated prefixes to be assigned to the network interface. You cannot use this
                    option if you use the Ipv4PrefixCount option.
                    * Ipv4Prefix (str, optional): The IPv4 prefix. For information, see  Assigning prefixes to Amazon EC2 network interfaces in
                        the Amazon Elastic Compute Cloud User Guide.
                * Ipv4PrefixCount (int, optional): The number of IPv4 delegated prefixes to be automatically assigned to the network interface. You
                    cannot use this option if you use the Ipv4Prefix option.
                * Ipv6Prefixes (List[Dict[str, Any]], optional): One or more IPv6 delegated prefixes to be assigned to the network interface. You cannot use this
                    option if you use the Ipv6PrefixCount option.
                    * Ipv6Prefix (str, optional): The IPv6 prefix.
                * Ipv6PrefixCount (int, optional): The number of IPv6 delegated prefixes to be automatically assigned to the network interface. You
                    cannot use this option if you use the Ipv6Prefix option.
            * Placement (Dict[str, Any], optional): The placement information for the instance.
                * AvailabilityZone (str, optional): The Availability Zone. [Spot Fleet only] To specify multiple Availability Zones, separate them
                    using commas; for example, "us-west-2a, us-west-2b".
                * GroupName (str, optional): The name of the placement group.
                * Tenancy (str, optional): The tenancy of the instance (if the instance is running in a VPC). An instance with a tenancy of
                    dedicated runs on single-tenant hardware. The host tenancy is not supported for Spot Instances.
            * RamdiskId (str, optional): The ID of the RAM disk.
            * SubnetId (str, optional): The ID of the subnet in which to launch the instance.
            * UserData (str, optional): The Base64-encoded user data for the instance. User data is limited to 16 KB.
        spot_price(str): The maximum price per hour that you are willing to pay for a Spot Instance.
            The default is the On-Demand price.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the spot instance.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        type(str):
            The Spot Instance request type.
            Default: one-time
            Valid Values: one-time | persistent
        valid_from(str):
            The start date of the request. If this is a one-time request,
            the request becomes active at this date and time and remains
            active until all instances launch, the request expires, or
            the request is canceled. If the request is persistent, the
            request becomes active at this date and time and remains
            active until it expires or is canceled.

            The specified start date and time cannot be equal to the current
            date and time. You must specify a start date and time that
            occurs after the current date and time.
        valid_until(str):
            The end date of the request, in UTC format (YYYY-MM-DDTHH:MM:SSZ).

            For a persistent request, the request remains active until the ValidUntil
            date and time is reached. Otherwise, the request remains active until you cancel it.

            For a one-time request, the request remains active until all instances launch,
            the request is canceled, or the ValidUntil date and time is reached.
            By default, the request is valid for 7 days from the date the request
            was created.

            Type: Timestamp

            Required: No

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.ec2.spot_instance.present:
                - name: value

        .. code-block:: sls

            resource_is_present:
              aws.ec2.spot_instance.present:
                instance_count: 5
                launch_specification:
                    ImageId: ami-fce3c696
                    KeyName: awskey.pem
                    SecurityGroups: ['sg-709f8709']
                    InstanceType: m4.large
                    Placement:
                        AvailabilityZone: us-east-1a
                    BlockDeviceMappings:
                            Ebs:
                                SnapshotId: snap-f70deff0
                                VolumeSize: 100
                                DeleteOnTermination: True
                                VolumeType: gp2
                                Iops: 300
                                Encrypted: False
                    EbsOptimized: True
                    Monitoring:
                        Enabled: True
                    SecurityGroupIds: sg-709f8709
                spot_price: 0.03
                type: one-time
    """
    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "name", "resource_id", "kwargs")
    }
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    for key in kwargs:
        result["comment"] += [f"Not explicitly supported keyword argument: '{key}'"]

    # Get the resource_id from ESM
    if resource_id is None:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if resource_id:
        # Assume that the spot_instance_request already exists since we have a resource_id
        result["old_state"] = await hub.tool.aws.ec2.spot_instance_request.get(
            ctx, resource_id=resource_id
        )

        if not result["old_state"]:
            result["comment"] += [
                f"Could not find spot_instance_request for '{name}' with existing id '{resource_id}'"
            ]
            return result
        result["comment"] += [f"Spot Instance Request '{name}' already exists"]
    else:
        if ctx.test:
            result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
            result["comment"] += [f"Would create aws.ec2.instance '{name}'"]
            return result
        if isinstance(tags, List):
            tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        tag_specification = (
            [
                {
                    "ResourceType": "instance",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None
        )
        # Although it's not specified in the API documentation,
        # apparently 'SecurityGroups' parameter requires the
        # names of the security groups, not the IDs, inside the
        # LaunchSpecification dict
        create_ret = await hub.exec.boto3.client.ec2.request_spot_instances(
            ctx,
            DryRun=False,
            AvailabilityZoneGroup=availability_zone_group,
            InstanceCount=instance_count,
            LaunchGroup=launch_group,
            LaunchSpecification=launch_specification,
            SpotPrice=spot_price,
            Type=type,
            ValidFrom=valid_from,
            ValidUntil=valid_until,
            TagSpecifications=tag_specification,
            InstanceInterruptionBehavior=instance_interruption_behavior,
        )
        result["result"] &= create_ret.result
        if not create_ret:
            result["comment"] += [create_ret.comment]
            return result

        result["comment"] += [f"Created '{name}'"]
        resource_id = create_ret.ret["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    # Update the spot_instance request now that it exists
    current_state = await hub.tool.aws.ec2.spot_instance_request.get(
        ctx, resource_id=resource_id
    )
    changes_made = False
    for attribute, new_value in desired_state.items():
        if new_value is None:
            # No value has been explicitly given, leave this parameter alone
            continue
        if attribute in (
            "availability_zone_group",
            "instance_count",
            "instance_interruption_behavior",
            "valid_from",
            "valid_until",
        ):
            result["comment"] += [
                f"'{attribute}' is only available on creation of a new spot instance request"
            ]
            continue
        if current_state.get(attribute) != new_value:
            changes_made = True
            if ctx.test:
                result["comment"] += [
                    f"Would update aws.ec2.instance '{name}': {attribute}"
                ]
                continue
            # TODO Update the spot instance request to match the changed parameters
            result["comment"] += [
                f"Currently unable to update aws.ec2.spot_instance_request attribute: {attribute}"
            ]

    # Get the final state of the resource
    if changes_made:
        if ctx.test:
            result["new_state"] = hub.tool.aws.ec2.spot_instance_request.test(
                **desired_state
            )
        else:
            result["new_state"] = await hub.tool.aws.ec2.spot_instance_request.get(
                ctx, resource_id=resource_id
            )
    else:
        result["new_state"] = current_state

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Cancels one or more Spot Instance requests.

    Args:
        hub:
        ctx:
        name(Text): An Idem name of the resource.
        resource_id(Text): The spot instance request id

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.spot_instance.absent:
                - name: value
    """
    result = dict(
        comment=[], old_state=ctx.old_state, new_state=None, name=name, result=True
    )

    # Get the resource_id from ESM
    if resource_id is None:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # If there still is no resource_id, the instance is gone
    if not resource_id:
        result["comment"] += [f"'{name}' already cancelled"]
        return result

    if ctx.test:
        result["comment"] += [f"Would cancel aws.ec2.spot_instance_request '{name}'"]
        return result

    ret = await hub.exec.boto3.client.ec2.cancel_spot_instance_requests(
        ctx,
        DryRun=False,
        SpotInstanceRequestIds=[resource_id],
    )

    result["result"] &= ret["result"]
    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result
    result["comment"] += [f"cancelled spot instance request '{name}'"]

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[Text, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.spot_instance_request
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(ctx)

    if not ret:
        hub.log.debug(f"Could not describe spot instance requests: {ret.comment}")
        return {}

    spot_instance_requests = (
        await hub.tool.aws.ec2.spot_instance_request.convert_to_present(ret.ret)
    )

    for resource_id, present_state in spot_instance_requests.items():
        result[resource_id] = {
            "aws.ec2.spot_instance_request.present": [
                {k: v} for k, v in present_state.items()
            ]
        }

    return result


async def search(hub, ctx, name: str, filters: List = None, resource_id: str = None):
    """
    Use an un-managed Instance as a data-source. Supply one of the inputs as the filter.

    Args:
        hub:
        ctx:
        name(Text): The name of the Idem state.
        resource_id(Text, optional): AWS VPC id to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key.
        A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instances

    Request Syntax:
        [Idem-state-name]:
          aws.ec2.spot_instance_request.search:
          - resource_id: 'Text'
          - filters:
            - name: 'string'
              values: 'list'
            - name: 'string'
              values: 'list'

        Examples:

            my-unmanaged-spot_instance_request:
              aws.ec2.spot_instance_request.search:
                - resource_id: value
    """
    result = dict(
        comment=[], old_state=ctx.old_state, new_state=None, name=name, result=True
    )

    # Get the resource_id from ESM if this resource used to be managed by "present"
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # Perform validation on the parameters
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    result["result"] &= syntax_validation["result"]
    if not result["result"]:
        result["comment"].append(syntax_validation["comment"])
        return result

    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    # Get all instances that match the given filters
    ret = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        ctx,
        Filters=boto3_filter,
        SpotInstanceRequestIds=[resource_id] if resource_id else None,
    )
    result["result"] &= ret.result
    if not ret:
        result["comment"].append(ret.comment)
        return result

    # Convert the described instances to the present state format
    spot_instance_requests = (
        await hub.tool.aws.ec2.spot_instance_request.convert_to_present(ret.ret)
    )

    # Check for null results
    if not spot_instance_requests:
        result["result"] = False
        result["comment"] += [
            f"Unable to find an aws.ec2.spot_instance_request for '{name}' that matched the given filters"
        ]
        return result

    # Get the first spot_instance_request from the results
    spot_instance_request_id = next(iter(spot_instance_requests.keys()))

    # Add a comment if there were multiple results
    if len(spot_instance_requests) > 1:
        result["comment"] += [
            f"More than one aws.ec2.spot_instance_request resource was found. Use resource '{spot_instance_request_id}'"
        ]

    # Return both old_state and new_state together
    result["new_state"] = spot_instance_requests[spot_instance_request_id]

    return result



import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
    waiter=None,
    waiter_args: Dict[str, Any] = None,
):
    """

    Args:
        hub:
        ctx:
        resource_arn(Text): aws elasticache arn
        old_tags(Dict[str, Any]): dict of old tags in the format of {tag-key: tag-value}
        new_tags(Dict[str, Any]): dict of new tags in the format of {tag-key: tag-value}. If this value is None, the function will do no operation on tags.
        waiter: waiter function to wait for the resource to be in available status after update.
        waiter_args(Dict[str, Any]): arguments provided for waiter function to execute.

    Returns:
        {"result": True|False, "comment": "A message tuple", "ret": dict of updated tags}

    """

    result = dict(comment=(), result=True, ret=None)

    tags_to_add = {}
    tags_to_delete = {}
    if new_tags is not None:
        tags_to_delete, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    if (not tags_to_delete) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if not ctx.get("test", False) and tags_to_delete:
        delete_tag_resp = (
            await hub.exec.boto3.client.elasticache.remove_tags_from_resource(
                ctx, ResourceName=resource_arn, TagKeys=list(tags_to_delete.keys())
            )
        )
        if not delete_tag_resp["result"]:
            hub.log.debug(
                f"Could not delete tags {tags_to_delete} for resource: '{resource_arn}' due to the error: {delete_tag_resp['comment']}"
            )
            result["comment"] = delete_tag_resp["comment"]
            result["result"] = False
            return result

        if waiter:
            # Optional waiter to wait for the resource to be in available status after updating the tags as the
            # resource might be in modifying status
            waiter_ret = await waiter(**waiter_args)
            if not waiter_ret["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + waiter_ret["comment"]

    if not ctx.get("test", False) and tags_to_add:
        create_tag_resp = await hub.exec.boto3.client.elasticache.add_tags_to_resource(
            ctx,
            ResourceName=resource_arn,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
        )
        if not create_tag_resp["result"]:
            hub.log.debug(
                f"Could not create tags {tags_to_add} for resource: '{resource_arn}' due to the error: {create_tag_resp['comment']}"
            )
            result["comment"] = create_tag_resp["comment"]
            result["result"] = False
            return result

        if waiter:
            # Optional waiter to wait for the resource to be in available status after updating the tags as the
            # resource might be in modifying status
            waiter_ret = await waiter(**waiter_args)
            if not waiter_ret["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + waiter_ret["comment"]
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_delete, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_delete, tags_to_add=tags_to_add
        )
    return result


def get_updated_cache_parameter_group(hub, old_paramaters, parameter_name_values):

    r"""
    Checks if the input ParameterNamesValues needs to be updated and returns a list of parameters that needs to be modified.

    Args:

        old_parameters: The detailed parameter list for a particular cache parameter group.
        parameter_name_values: Parameters Values as needed by the user.

    Returns:
        [
            {
                "ParameterName": "Text",
                "ParameterValue": "Text"
            },
        ]
    """
    final_parameters = []
    old_parameter_map = {
        params.get("ParameterName"): params for params in old_paramaters
    }
    if parameter_name_values:
        for new_parameters in parameter_name_values:
            if new_parameters.get("ParameterName") in old_parameter_map:
                if not new_parameters.get("ParameterValue") == old_parameter_map.get(
                    new_parameters.get("ParameterName")
                ).get("ParameterValue"):
                    final_parameters.append(new_parameters)
    return final_parameters


async def update_replication_group(
    hub,
    ctx,
    name,
    resource_id: str,
    old_state: Dict[str, Any] = None,
    plan_state: Dict[str, Any] = None,
    replication_group_description: str = None,
    primary_cluster_id: str = None,
    snapshotting_cluster_id: str = None,
    automatic_failover_enabled: bool = None,
    multi_az_enabled: bool = None,
    node_group_id: str = None,
    cache_security_group_names: List[str] = None,
    security_group_ids: List[str] = None,
    preferred_maintenance_window: str = None,
    notification_topic_arn: str = None,
    cache_parameter_group_name: str = None,
    notification_topic_status: str = None,
    apply_immediately: bool = None,
    engine_version: str = None,
    auto_minor_version_upgrade: bool = None,
    snapshot_retention_limit: int = None,
    snapshot_window: str = None,
    cache_node_type: str = None,
    auth_token: str = None,
    auto_token_update_strategy: str = None,
    user_group_ids_to_add: List[str] = None,
    user_group_ids_to_remove: List[str] = None,
    remove_user_groups: bool = None,
    log_delivery_configurations: List[
        make_dataclass(
            "LogDeliveryConfigurationRequest",
            [
                ("LogType", str, field(default=None)),
                ("DestinationType", str, field(default=None)),
                (
                    "DestinationDetails",
                    make_dataclass(
                        "DestinationDetails",
                        [
                            (
                                "CloudWatchLogsDetails",
                                make_dataclass(
                                    "CloudWatchLogsDestinationDetails",
                                    [("LogGroup", str, field(default=None))],
                                ),
                                field(default=None),
                            ),
                            (
                                "KinesisFirehoseDetails",
                                make_dataclass(
                                    "KinesisFirehoseDestinationDetails",
                                    [("DeliveryStream", str, field(default=None))],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                ("LogFormat", str, field(default=None)),
                ("Enabled", bool, field(default=None)),
            ],
        )
    ] = None,
):
    """
    Modifies the settings for a replication group.
    - Scaling for Amazon ElastiCache for Redis (cluster mode enabled) in the ElastiCache User Guide
    - ModifyReplicationGroupShardConfiguration in the ElastiCache API Reference
    Note: This operation is valid for Redis only.

    Args:
        name(Text): An Idem name of the resource. It is used as ReplicationGroupID during resource creation
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        replication_group_description(Text): A user-created description for the replication group.
        primary_cluster_id(Text, optional): The identifier of the cluster that serves as the primary for this replication group. This
            cluster must already exist and have a status of available. This parameter is not required if
            NumCacheClusters, NumNodeGroups, or ReplicasPerNodeGroup is specified. Defaults to None.
        snapshotting_cluster_id (Text, optional): The cluster ID that is used as the daily snapshot source for the replication group. This parameter
                    cannot be set for Redis (cluster mode enabled) replication groups.
        automatic_failover_enabled(bool, optional): Specifies whether a read-only replica is automatically promoted to read/write primary if the
            existing primary fails.  AutomaticFailoverEnabled must be enabled for Redis (cluster mode
            enabled) replication groups. Default: false. Defaults to None.
        multi_az_enabled(bool, optional): A flag indicating if you have Multi-AZ enabled to enhance fault tolerance. For more information,
            see Minimizing Downtime: Multi-AZ. Defaults to None.
        node_group_id(Text, optional): Deprecated. This parameter is not used.
        notification_topic_status(bool, optional): The Amazon Resource Name (ARN) of the Amazon SNS topic to which notifications are sent.
        apply_immediately(bool, optional): If true , this parameter causes the modifications in this request and any pending modifications to be applied,
            asynchronously and as soon as possible, regardless of the PreferredMaintenanceWindow setting for the replication group.
            If false , changes to the nodes in the replication group are applied on the next maintenance reboot, or the next failure reboot,
            whichever occurs first.
            Valid values: true | false
            Default: false
        cache_security_group_names(List[str], optional): A list of cache security group names to associate with this replication group. Defaults to None.
        security_group_ids(List[str], optional): One or more Amazon VPC security groups associated with this replication group. Use this
            parameter only when you are creating a replication group in an Amazon Virtual Private Cloud
            (Amazon VPC). Defaults to None.
        preferred_maintenance_window(Text, optional): Specifies the weekly time range during which maintenance on the cluster is performed. It is
            specified as a range in the format ddd:hh24:mi-ddd:hh24:mi (24H Clock UTC). The minimum
            maintenance window is a 60 minute period. Valid values for ddd are: Specifies the weekly time
            range during which maintenance on the cluster is performed. It is specified as a range in the
            format ddd:hh24:mi-ddd:hh24:mi (24H Clock UTC). The minimum maintenance window is a 60 minute
            period. Valid values for ddd are:    sun     mon     tue     wed     thu     fri     sat
            Example: sun:23:00-mon:01:30. Defaults to None.
        engine_version(Text, optional): The version number of the cache engine to be used for the clusters in this replication group. To
            view the supported cache engine versions, use the DescribeCacheEngineVersions operation.
            Important: You can upgrade to a newer engine version (see Selecting a Cache Engine and Version)
            in the ElastiCache User Guide, but you cannot downgrade to an earlier engine version. If you
            want to use an earlier engine version, you must delete the existing cluster or replication group
            and create it anew with the earlier engine version. Defaults to None.
        auto_minor_version_upgrade(bool, optional):  If you are running Redis engine version 6.0 or later, set this parameter to yes if you want to
            opt-in to the next auto minor version upgrade campaign. This parameter is disabled for previous
            versions. . Defaults to None.
        snapshot_retention_limit(int, optional): The number of days for which ElastiCache retains automatic snapshots before deleting them. For
            example, if you set SnapshotRetentionLimit to 5, a snapshot that was taken today is retained for
            5 days before being deleted. Default: 0 (i.e., automatic backups are disabled for this cluster). Defaults to None.
        snapshot_window(Text, optional): The daily time range (in UTC) during which ElastiCache begins taking a daily snapshot of your
            node group (shard). Example: 05:00-09:00  If you do not specify this parameter, ElastiCache
            automatically chooses an appropriate time range. Defaults to None.
        auth_token(Text, optional):  Reserved parameter. The password used to access a password protected server.  AuthToken can be
            specified only on replication groups where TransitEncryptionEnabled is true.  For HIPAA
            compliance, you must specify TransitEncryptionEnabled as true, an AuthToken, and a
            CacheSubnetGroup.  Password constraints:   Must be only printable ASCII characters.   Must be at
            least 16 characters and no more than 128 characters in length.   The only permitted printable
            special characters are !, &, #, $, ^, <, >, and -. Other printable special characters cannot be
            used in the AUTH token.   For more information, see AUTH password at
            http://redis.io/commands/AUTH. Defaults to None.
        cache_node_type(Text, optional): The compute and memory capacity of the nodes in the node group (shard). The following node types
            are supported by ElastiCache. Generally speaking, the current generation types provide more
            memory and computational power at lower cost when compared to their equivalent previous
            generation counterparts.
        auto_token_update_strategy(Text, optional): Specifies the strategy to use to update the AUTH token. This parameter must be specified with
                the auth-token parameter. Possible values:
                Rotate
                Set
                For more information, see Authenticating Users with Redis AUTH
        user_group_ids_to_add(List[str], optional): The ID of the user group you are associating with the replication group.
        user_group_ids_to_remove(List[str], optional): The ID of the user group to disassociate from the replication group, meaning the users in
                the group no longer can access the replication group.
        remove_user_groups(bool, optional): Removes the user group associated with this replication group.
        log_delivery_configurations(List[Dict], optional): Specifies the destination, format and type of the logs. Defaults to None.
            * LogType (Text, optional): Refers to slow-log.
            * DestinationType (Text, optional): Specify either cloudwatch-logs or kinesis-firehose as the destination type.
            * DestinationDetails (Text, optional): Configuration details of either a CloudWatch Logs destination or Kinesis Data Firehose
            destination.
                * CloudWatchLogsDetails (Dict, optional): The configuration details of the CloudWatch Logs destination.
                    * LogGroup (Text, optional): The name of the CloudWatch Logs log group.
                * KinesisFirehoseDetails (Dict, optional): The configuration details of the Kinesis Data Firehose destination.
                    * DeliveryStream (str, optional): The name of the Kinesis Data Firehose delivery stream.
            * LogFormat (Text, optional): Specifies either JSON or TEXT
            * Enabled (bool, optional): Specify if log delivery is enabled. Default true.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    params_to_modify = {}
    modify_params = OrderedDict(
        {
            "ReplicationGroupDescription": "replication_group_description",
            "PrimaryClusterId": "primary_cluster_id",
            "SnapshottingClusterId": "snapshotting_cluster_id",
            "AutomaticFailoverEnabled": "automatic_failover_enabled",
            "MultiAZEnabled": "multi_az_enabled",
            "NodeGroupId": "node_group_id",
            "CacheSecurityGroupNames": "cache_security_group_names",
            "SecurityGroupIds": "security_group_ids",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "NotificationTopicArn": "notification_topic_arn",
            "CacheParameterGroupName": "cache_parameter_group_name",
            "NotificationTopicStatus": "notification_topic_status",
            "ApplyImmediately": "apply_immediately",
            "EngineVersion": "engine_version",
            "AutoMinorVersionUpgrade": "auto_minor_version_upgrade",
            "SnapshotRetentionLimit": "snapshot_retention_limit",
            "SnapshotWindow": "snapshot_window",
            "CacheNodeType": "cache_node_type",
            "AuthToken": "auth_token",
            "AuthTokenUpdateStrategy": "auto_token_update_strategy",
            "UserGroupIdsToAdd": "user_group_ids_to_add",
            "UserGroupIdsToRemove": "user_group_ids_to_remove",
            "RemoveUserGroups": "remove_user_groups",
            "LogDeliveryConfigurations": "log_delivery_configurations",
        }
    )
    for parameter_raw, parameter_present in modify_params.items():
        # Add to modify list only if parameter is changed
        input_parameter_value = locals()[parameter_present]
        if (
            input_parameter_value is not None
            and old_state.get(parameter_present) != input_parameter_value
        ):
            params_to_modify[parameter_raw] = input_parameter_value
    if params_to_modify:
        if ctx.get("test", False):
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
            for key, value in params_to_modify.items():
                plan_state[modify_params.get(key)] = value
        else:
            modify_ret = (
                await hub.exec.boto3.client.elasticache.modify_replication_group(
                    ctx, ReplicationGroupId=resource_id, **params_to_modify
                )
            )
            result["result"] = result["result"] and modify_ret["result"]
            result["comment"] += hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
            if not modify_ret["result"]:
                result["comment"] = result["comment"] + modify_ret["comment"]
                result["result"] = False
                return result
    return result


async def replication_group_waiter(
    hub, ctx, name: str, resource_id: str, timeout: Dict, operation_type: str
):
    """

    Waiter to wait for the replication group to become active.

        Args:
           name(Text): An Idem name of the resource. It is used as ReplicationGroupID during resource creation
           resource_id(Text): Replication group id to Identify the resource
           timeout(Dict, optional): Timeout configuration for creating or updating replication group.
            * create (Dict) -- Timeout configuration for creating replication group
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
            * update (Text) -- Timeout configuration for updating replication group
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
           operation_type(Text): create or update operation

        Returns:
            Dict[str, Any]
    """

    result = dict(comment=[], result=True, ret={})
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=15,
        default_max_attempts=40,
        timeout_config=timeout.get(operation_type) if timeout else None,
    )
    hub.log.debug(
        f"Waiting on {operation_type} aws.elasticache.replication_group '{name}'"
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "elasticache",
            "replication_group_available",
            ReplicationGroupId=resource_id,
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] += list(str(e))
        result["result"] = False
    return result



from typing import Any
from typing import Dict

from dict_tools import data


async def update(
    hub,
    ctx,
    resource_id: str,
    raw_resource: Dict[str, Any],
    access_control_policy: Dict[str, None],
) -> Dict[str, Any]:
    """Updates an AWS Bucket ACL resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id(str): ACL Policy name identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        access_control_policy(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    acl_policy_diff = data.recursive_diff(
        access_control_policy, raw_resource, ignore_order=True
    )

    if acl_policy_diff:
        result["ret"] = {}

        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.s3.put_bucket_acl(
                ctx,
                Bucket=resource_id,
                AccessControlPolicy=access_control_policy,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.s3.bucket_acl", name=resource_id
            )
        result["ret"]["access_control_policy"] = access_control_policy

    return result



import pathlib

import boto3.session
import botocore.exceptions
from dict_tools.data import NamespaceDict

try:
    import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx, directory)

    session = boto3.session.Session()
    services = hub.OPT.pop_create.services or session.get_available_services()
    ctx.servers = [None]

    # We already have an acct plugin
    ctx.has_acct_plugin = False
    ctx.service_name = "aws_auto"

    # Initialize cloud spec
    ctx.cloud_spec = NamespaceDict(
        api_version="",
        project_name=ctx.project_name,
        service_name=ctx.service_name,
        request_format={
            "present": hub.pop_create.aws.template.PRESENT_REQUEST_FORMAT,
            "absent": hub.pop_create.aws.template.ABSENT_REQUEST_FORMAT,
            "describe": hub.pop_create.aws.template.DESCRIBE_REQUEST_FORMAT,
        },
        plugins={},
    )

    # This takes a while because we are making http calls to aws
    for service in tqdm.tqdm(services):
        try:
            plugins = hub.pop_create.aws.plugin.parse(session, service)
        except botocore.exceptions.UnknownServiceError as err:
            plugins = {}
            hub.log.error(f"{err.__class__.__name__}: {err}")

        ctx.cloud_spec.plugins = plugins

        # Generate the exec modules for this specific service
        hub.cloudspec.init.run(
            ctx,
            directory,
            create_plugins=["state_modules"],
        )

    ctx.cloud_spec.plugins = {}
    return ctx



"""Exec module for managing IAM Groups."""


async def get(hub, ctx, name: str, resource_id: str):
    """Get an IAM group from AWS.

    Provide group name as input.

    Args:
        hub: required for functions in hub.
        ctx: context.
        name(str): An Idem name of the IAM group.
        resource_id(str, optional): AWS IAM group name.

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.iam.get_group(ctx, GroupName=resource_id)
    if not ret["result"]:
        if "NoSuchEntity" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.group", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    result["ret"] = hub.tool.aws.iam.conversion_utils.convert_raw_group_to_present(
        raw_resource=ret["ret"]["Group"], idem_resource_name=resource_id
    )
    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_name: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_name: aws cloudwatch resource name
        old_tags: Dict of existing tags
        new_tags: Dict of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": dict of updated tags}

    """
    result = dict(comment=(), result=True, ret={})
    tags_to_add = {}
    tags_to_delete = {}
    if new_tags is not None:
        tags_to_delete, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    if (not tags_to_delete) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if not ctx.get("test", False) and tags_to_delete:
        delete_tag_resp = await hub.exec.boto3.client.logs.untag_log_group(
            ctx, logGroupName=resource_name, tags=list(tags_to_delete.keys())
        )
        if not delete_tag_resp["result"]:
            result["comment"] = delete_tag_resp["comment"]
            result["result"] = False
            return result
    if not ctx.get("test", False) and tags_to_add:
        create_tag_resp = await hub.exec.boto3.client.logs.tag_log_group(
            ctx, logGroupName=resource_name, tags=tags_to_add
        )
        if not create_tag_resp["result"]:
            result["comment"] = result["comment"] + create_tag_resp["comment"]
            result["result"] = False
            return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_delete, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_delete, tags_to_add=tags_to_add
        )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_method_to_present(
    hub,
    raw_resource: Any,
    resource_id: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway Method to a common idem present state.

    Args:
        resource_id(string): Idem Resource ID that is generated once the resource is created.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(str, optional): The idem name of the resource.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "httpMethod": "http_method",
            "authorizationType": "authorization_type",
            "apiKeyRequired": "api_key_required",
        }
    )

    resource_translated = {
        "resource_id": resource_id,
        "rest_api_id": resource_id.split("-")[0],
        "parent_resource_id": resource_id.split("-")[1],
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if idem_resource_name:
            resource_translated["name"] = idem_resource_name
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



"""This module defines the present, absent and describe functionality for AWS neptune db cluster parameter group."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

RESOURCE_TYPE = "aws.neptune.db_cluster_parameter_group"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    db_parameter_group_family: str,
    description: str,
    parameters: List[
        make_dataclass(
            """DB Parameter group configuration.""" "Parameter",
            [
                ("ParameterName", str),
                ("ParameterValue", str),
                ("ApplyMethod", str, field(default=None)),
                ("Description", str, field(default=None)),
                ("Source", str, field(default=None)),
                ("ApplyType", str, field(default=None)),
                ("DataType", str, field(default=None)),
                ("AllowedValues", str, field(default=None)),
                ("MinimumEngineVersion", str, field(default=None)),
                ("SupportedEngineModes", List[str], field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Creates a new DB cluster parameter group.

    Parameters in a DB cluster parameter group apply to all of the
    instances in a DB cluster.  A DB cluster parameter group is initially created with the default parameters for
    the database engine used by instances in the DB cluster if parameters is not specified. When you associate a new

    .. warning::
        DB cluster parameter group with a running DB cluster, you need to reboot the DB instances in the DB cluster
        without failover for the new DB cluster parameter group and associated settings to take effect.  After you
        create a DB cluster parameter group, you should wait at least 5 minutes before creating your first DB cluster
        that uses that DB cluster parameter group as the default parameter group. This allows Amazon Neptune to fully
        complete the create action before the DB cluster parameter group is used as the default for a new DB cluster.
        This is especially important for parameters that are critical when creating the default database for a DB
        cluster, such as the character set for the default database defined by the character_set_database parameter. You
        can use the Parameter Groups option of the Amazon Neptune console or the DescribeDBClusterParameters command to
        verify that your DB cluster parameter group has been created or modified.

    Args:
        name(str):
            DBClusterParameterGroupName parameter on the resource. This value is stored as a lowercase string.
            This is also used as idem name.

        resource_id(str, Optional):
            DBClusterParameterGroupName of the resource. Defaults to None.

        db_parameter_group_family(str):
            The DB cluster parameter group family name. A DB cluster parameter group can be
            associated with one and only one DB cluster parameter group family, and can be applied only to a DB cluster
            running a database engine and engine version compatible with that DB cluster parameter group
            family.

        description(str):
            The description for the DB cluster parameter group. This is required field.

        parameters(list, Optional): parameters in the DB cluster parameter group. Defaults to None.
            These parameters can only be modified.
            To modify more than one parameter, submit a list of the following: ParameterName , ParameterValue , and
            ApplyMethod . A maximum of 20 parameters can be modified in a single request.

            .. note::
                Changes to dynamic parameters are applied immediately. Changes to static parameters require a reboot
                without failover to the DB cluster associated with the parameter group before the change can take
                effect.

            Below are the attributes available for parameters:

            Type: object dict(Describes a parameter)

            * ParameterName (str):
                Specifies the name of the parameter.
            * ParameterValue (str):
                Specifies the value of the parameter.
            * Description (str, Optional):
                provides a description of the parameter.
            * Source (str, Optional):
                Indicates the source of the parameter value.
            * ApplyType (str, Optional):
                Specifies the engine specific parameters type.
            * DataType (str, Optional):
                Specifies the valid data type for the parameter.
            * AllowedValues (str, Optional):
                Specifies the valid range of values for the parameter.
            * MinimumEngineVersion (str, Optional):
                The earliest engine version to which the parameter can apply.
            * ApplyMethod (str, Optional):
                Indicates when to apply parameter updates.
            * SupportedEngineModes (list, Optional):
                The valid DB engine modes.

        tags(dict, Optional):
            The tags to be assigned to the new DB cluster parameter group. Defaults to None.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_neptune_db_cluster_parameter_group]:
              aws.neptune.db_cluster_parameter_group.present:
              - db_cluster_parameter_group_arn: 'string'
                db_parameter_group_family: 'string'
                description: 'string'
                name: 'string'
                parameters:
                - AllowedValues: 'string'
                  ApplyMethod: 'string'
                  ApplyType: 'string'
                  DataType: 'string'
                  Description: 'string'
                  IsModifiable: 'string'
                  ParameterName: 'string'
                  Source: 'string'


    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            default.aurora5.6:
              aws.neptune.db_cluster_parameter_group.present:
              - db_cluster_parameter_group_arn: arn:aws:rds:ca-central-1:746014882121:cluster-pg:default.aurora5.6
                db_parameter_group_family: aurora5.6
                description: Default cluster parameter group for aurora5.6
                name: default.aurora5.6
                parameters:
                - AllowedValues: 0,1
                  ApplyMethod: pending-reboot
                  ApplyType: static
                  DataType: boolean
                  Description: Controls whether user-defined functions that have only an xxx symbol
                    for the main function can be loaded
                  IsModifiable: false
                  ParameterName: allow-suspicious-udfs
                  Source: engine-default
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    desired_state = {
        "name": name,
        "resource_id": name,
        "db_parameter_group_family": db_parameter_group_family,
        "description": description,
        "parameters": parameters,
        "tags": tags,
    }
    # set a flag to keep track if the resource is newly created
    is_created = False

    if not resource_id:
        # if there is no resource_id, it is expected that we create the resource
        # if this is test run, just generate the new state and set in result
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            resource_id = desired_state["resource_id"]
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        else:
            # if this is not a test run, create the real resource
            create_db_cluster_parameter_group_result = (
                await hub.exec.boto3.client.neptune.create_db_cluster_parameter_group(
                    ctx,
                    DBClusterParameterGroupName=name,
                    DBParameterGroupFamily=db_parameter_group_family,
                    Description=description,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
            )
            # if create failed, fail the request
            if not create_db_cluster_parameter_group_result["result"]:
                result["comment"] = create_db_cluster_parameter_group_result["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
            resource_id = create_db_cluster_parameter_group_result["ret"][
                "DBClusterParameterGroup"
            ]["DBClusterParameterGroupName"]
        is_created = True

    # The argument "parameters" can only be modified as per AWS docs, so if user specifies
    # this during create, idem will first create the db_cluster_parameter_group and then modify its parameters
    # P.S. There is no way to create these parameters as per AWS doc.
    update_db_cluster_parameter_group_result = (
        await hub.tool.aws.neptune.db_cluster_parameter_group.update(
            ctx,
            name=name,
            resource_id=resource_id,
            desired_state=desired_state,
            is_created=is_created,
        )
    )
    if not (
        update_db_cluster_parameter_group_result
        and update_db_cluster_parameter_group_result["result"]
        and update_db_cluster_parameter_group_result["new_state"]
    ):
        result["result"] = False
        result["comment"] = update_db_cluster_parameter_group_result["comment"]
        return result

    result["old_state"] = update_db_cluster_parameter_group_result["old_state"]
    result["new_state"] = update_db_cluster_parameter_group_result["new_state"]
    # update the result comment if the resource is not being created
    if not is_created:
        result["comment"] = update_db_cluster_parameter_group_result["comment"]

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes a specified DB cluster parameter group.

    The DB cluster parameter group to be deleted can't be associated
    with any DB clusters.

    Args:
        name(str):
            DBClusterParameterGroupName parameter on the resource.
            This is also used as idem name.

        resource_id(str, Optional):
            DBClusterParameterGroupName of the resource. Defaults to None.
            Idem automatically considers the resource being absent if resource_id is not specified.
            Constraints:   Must match the name of an existing DBClusterParameterGroupName.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_neptune_db_cluster_parameter_group]:
                aws.neptune.db_cluster_parameter_group.absent:
                    - name: 'string'
                    - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.neptune.db_cluster_parameter_group.absent:
                - name: 'test-db-clus-param-group'
                - resource_id: 'test-db-clus-param-group'
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type=RESOURCE_TYPE, name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.neptune.db_cluster_parameter_group.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        return result
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.neptune.delete_db_cluster_parameter_group(
            ctx,
            DBClusterParameterGroupName=resource_id,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns a list of DBClusterParameterGroup descriptions.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.neptune.db_cluster_parameter_group
    """
    result = {}
    ret = await hub.exec.aws.neptune.db_cluster_parameter_group.list(
        ctx, name="aws.neptune.db_cluster_parameter_group.describe"
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe {RESOURCE_TYPE} {ret['comment']}")
        return result

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "aws.neptune.db_cluster_parameter_group.present": [resource]
        }
    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """Update tags of AWS Dynamo DB resources

    Args:
        resource_arn:
            Identifies the Amazon DynamoDB resource to which tags should be added.
            This value is an Amazon Resource Name (ARN).

        old_tags:
            Dict of existing tags.

        new_tags:
            Dict of new tags.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": dict of updated tags}

    """

    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.dynamodb.untag_resource(
                ctx,
                ResourceArn=resource_arn,
                TagKeys=list(tags_to_remove.keys()),
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.dynamodb.tag_resource(
                ctx,
                ResourceArn=resource_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] += add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



"""State module for managing Amazon WAFv2 Regex Pattern Set."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    regular_expression_list: List[
        make_dataclass("Regex", [("RegexString", str, field(default=None))])
    ],
    resource_id: str = None,
    description: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """Creates a RegexPatternSet, which you reference in a RegexPatternSetReferenceStatement, to have WAF inspect a web
    request component for the specified patterns.

    Args:
        name(str): The name of the set. You cannot change the name after you create the set.
        resource_id(str, Optional): An identifier of the resource in the provider. Defaults to None.
        scope(str): Specifies whether this is for an Amazon CloudFront distribution or for a regional application. A
            regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API,
            or an AppSync GraphQL API.  To work with CloudFront, you must also specify the Region US East
            (N. Virginia) as follows:    CLI - Specify the Region when you use the CloudFront scope:
            --scope=CLOUDFRONT --region=us-east-1.    API and SDKs - For all calls, use the Region endpoint
            us-east-1.
        description(str, Optional): A description of the set that helps with identification. Defaults to None.
        regular_expression_list(list): Array of regular expression strings.
            * RegexString (str, Optional): The string representing the regular expression.
        tags(dict, Optional): Dict in the format of {tag-key: tag-value}

    Request Syntax:
       .. code-block:: sls

          [regex-pattern-resource-id]:
              aws.wafv2.regex_pattern_set.present:
                - name: "string"
                - scope: "string"
                - regular_expression_list:
                    - RegexString: "string"
                - tags:
                    scope: "string"
                    name: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            regex_pattern_set_name:
              aws.wafv2.regex_pattern_set.present:
                - name: regex_pattern_set_name
                - scope: CLOUDFRONT
                - regular_expression_list:
                    - RegexString: "idem-aws/$+"
                - tags:
                    scope: cloudfront-1
                    name: regex_pattern_set_name
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    is_invalid_scope_and_region = (
        hub.tool.aws.wafv2.regex_pattern_set.check_if_invalid_scope_and_region(
            ctx["acct"].get("region_name"), scope
        )
    )
    if is_invalid_scope_and_region:
        result["result"] = False
        result["comment"] = (is_invalid_scope_and_region,)
        return result
    if resource_id:
        before = await hub.exec.aws.wafv2.regex_pattern_set.get(
            ctx=ctx, name=name, resource_id=resource_id, scope=scope
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = await hub.tool.aws.wafv2.regex_pattern_set.update(
            ctx=ctx,
            current_state=result["old_state"],
            description=description,
            regular_expression_list=regular_expression_list,
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            for modified_param in update_ret["ret"]:
                plan_state[modified_param] = update_ret["ret"][modified_param]
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.wafv2.regex_pattern_set", name=name
            )

        # update tags
        if tags is not None and tags != result["old_state"].get("tags"):
            update_tags_ret = await hub.tool.aws.wafv2.tag.update_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            result["result"] = result["result"] and update_tags_ret["result"]
            resource_updated = resource_updated or bool(update_tags_ret["ret"])

            if ctx.get("test", False) and update_tags_ret["ret"]:
                plan_state["tags"] = update_tags_ret["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "scope": scope,
                    "regular_expression_list": regular_expression_list,
                    "description": description,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.wafv2.regex_pattern_set", name=name
            )
            return result
        ret = await hub.exec.boto3.client.wafv2.create_regex_pattern_set(
            ctx=ctx,
            Name=name,
            Scope=scope,
            Description=description,
            RegularExpressionList=regular_expression_list,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["Summary"]["Id"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.wafv2.regex_pattern_set", name=name
        )
    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.aws.wafv2.regex_pattern_set.get(
            ctx=ctx, name=name, resource_id=resource_id, scope=scope
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub, ctx, name: str, scope: str, resource_id: str = None
) -> Dict[str, Any]:
    """Deletes the specified RegexPatternSet.

    Args:
        name(str): The name of the set. You cannot change the name after you create the set.
        resource_id(str, Optional): A unique identifier for the set. This ID is returned in the responses
            to create and list commands. You provide it to operations like update and delete.
            Idem automatically considers this resource being absent if this field is not specified. Defaults to None.
        scope(str): Specifies whether this is for an Amazon CloudFront distribution or for a regional application. A
            regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API,
            or an AppSync GraphQL API.  To work with CloudFront, you must also specify the Region US East
            (N. Virginia) as follows:    CLI - Specify the Region when you use the CloudFront scope:
            --scope=CLOUDFRONT --region=us-east-1.    API and SDKs - For all calls, use the Region endpoint
            us-east-1.

    Request Syntax:
       .. code-block:: sls

          [regex-pattern-resource-id]:
              aws.wafv2.regex_pattern_set.absent:
                - name: "string"
                - scope: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            regex_pattern_set_name:
              aws.wafv2.regex_pattern_set.absent:
                - name: regex_pattern_set_name
                - resource_id: value
                - scope: CLOUDFRONT
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.wafv2.regex_pattern_set", name=name
        )
        return result
    before = await hub.exec.aws.wafv2.regex_pattern_set.get(
        ctx=ctx, name=name, resource_id=resource_id, scope=scope
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.wafv2.regex_pattern_set", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.wafv2.regex_pattern_set", name=name
        )
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.wafv2.delete_regex_pattern_set(
            ctx=ctx,
            Name=name,
            Scope=scope,
            Id=resource_id,
            LockToken=before["ret"]["LockToken"],
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.wafv2.regex_pattern_set", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the
        corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.wafv2.regex_pattern_set
    """

    result = {}
    # Two types are scopes are allowed if the region is us-east-1 else only REGIONAL is allowed
    scopes_allowed = (
        ["CLOUDFRONT", "REGIONAL"]
        if ctx["acct"].get("region_name") == "us-east-1"
        else ["REGIONAL"]
    )

    for scope in scopes_allowed:
        ret = await hub.exec.boto3.client.wafv2.list_regex_pattern_sets(
            ctx=ctx, Scope=scope
        )

        if not ret["result"]:
            hub.log.debug(
                f"Could not describe wafv2 regex_pattern_set for scope {scope} {ret['comment']}"
            )
            continue

        for regex_pattern_set in ret["ret"]["RegexPatternSets"]:
            get_regex_ret = await hub.exec.aws.wafv2.regex_pattern_set.get(
                ctx=ctx,
                name=regex_pattern_set["Name"],
                resource_id=regex_pattern_set["Id"],
                scope=scope,
            )
            if not get_regex_ret["result"]:
                hub.log.debug(
                    f"Could not describe wafv2 regex_pattern_set for scope {regex_pattern_set['Name']} {get_regex_ret['comment']}"
                )
                continue
            resource_translated = get_regex_ret["ret"]
            # LockToken is not needed in describe.
            if "LockToken" in resource_translated:
                resource_translated.pop("LockToken")
            result[resource_translated["resource_id"]] = {
                "aws.wafv2.regex_pattern_set.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
    return result



import contextlib
import os
import pathlib
import tempfile
from typing import Tuple

from Cryptodome.PublicKey import RSA


@contextlib.contextmanager
def verify(
    hub, *, ssh_public_key: str = None, ssh_private_key: str = None
) -> Tuple[str, str]:
    """
    Validate the ssh keypair, generate a keypair if necessary

    Args:
        hub:
        ssh_public_key(Text, Optional): A public ssh key or path to send to the instance
        ssh_private_key(Text, Optional): A private ssh key or path to send to the instance

    Yields:
        A tuple containing the bytes of the public ssh key and the file path of the private ssh key as a tuple

    Examples:
        # Create a new keypair
        with hub.tool.aws.ec2.instance.key_pair.verify() as key_pair:
            verified_public_key, private_key_file = key_pair

        # Validate an existing keypair
        ssh_public_key = "ssh-rsa XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX username@host"
        ssh_private_key = "~/.ssh/id_rsa"
        with hub.tool.aws.ec2.instance.key_pair.verify(ssh_public_key=ssh_public_key, ssh_private_key=ssh_private_key) as key_pair:
            verified_public_key, private_key_file = key_pair
    """
    if bool(ssh_public_key) is not bool(ssh_private_key):
        raise ValueError(
            "Both a private ssh key and a public ssh key must be specified"
        )

    # Create a keypair if none were passed
    if not (ssh_public_key and ssh_private_key):
        hub.log.debug(f"Creating an ssh keypair")
        key_pair = RSA.generate(2048)
        ssh_public_key = key_pair.publickey().export_key("OpenSSH")
        ssh_private_key = key_pair.export_key("PEM")

    # Verify the public key
    if isinstance(ssh_public_key, bytes):
        ssh_public_key = ssh_public_key.decode()
    public_key_path = pathlib.Path(ssh_public_key)
    # If the ssh_public_key was a path, then read the contents of that path into a variable
    if public_key_path.exists():
        ssh_public_key = public_key_path.read_text()

    # Verify the private key
    private_key_path = pathlib.Path(
        ssh_private_key.decode()
        if isinstance(ssh_private_key, bytes)
        else ssh_private_key
    )
    if private_key_path.exists():
        st_mode = private_key_path.stat().st_mode
        if st_mode != 0o600:
            raise PermissionError(
                f"Private key has invalid permissions, should be 0o600: {oct(st_mode)}"
            )
        yield private_key_path.name
    else:
        with tempfile.NamedTemporaryFile(
            mode="wb+", prefix="id-rsa-", suffix=".pem", delete=True
        ) as private_key_file:
            # Write private_key to file and chmod 400
            private_key_file.write(ssh_private_key)
            private_key_file.flush()
            os.chmod(private_key_file.name, mode=0o600)  # noqa
            yield ssh_public_key, private_key_file.name



"""Exec module for managing IAM Group Policy Attachment."""
from typing import Any
from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
    group: str,
    policy_arn: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Get managed policy attached to a group.

    Args:
        hub: required for functions in hub.

        ctx: context.

        name(str): An Idem name of the state for logging.

        group(str): The group name to attach the policy to.

        policy_arn(str, optional): Policy ARN.

        resource_id(str, optional): Policy ARN
    """
    result = dict(comment=[], result=True, ret=None)

    ret = await hub.exec.boto3.client.iam.list_attached_group_policies(
        ctx, GroupName=group
    )
    if not resource_id:
        resource_id = policy_arn

    if not ret["result"]:
        if "NoSuchEntity" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.group_policy_attachment", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    if ret["result"]:
        attached_group_policies_list = ret["ret"].get("AttachedPolicies")
        if attached_group_policies_list:
            policy_arn_list = [
                policy.get("PolicyArn") for policy in attached_group_policies_list
            ]
            if resource_id in policy_arn_list:
                result[
                    "ret"
                ] = hub.tool.aws.iam.conversion_utils.convert_raw_group_policy_attachment_to_present(
                    group=group, policy_arn=resource_id
                )
            else:
                result["comment"].append(
                    hub.tool.aws.comment_utils.get_empty_comment(
                        resource_type="aws.iam.group_policy_attachment", name=name
                    )
                )

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.associate_route_table
hub.exec.boto3.client.ec2.disassociate_route_table
hub.tool.boto3.resource.exec(resource, associate_with_subnet, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    route_table_id: str,
    resource_id: str = None,
    subnet_id: str = None,
    gateway_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Associates a subnet in your VPC or an internet gateway or virtual private gateway attached to your VPC with a
    route table in your VPC. This association causes traffic from the subnet or gateway to be routed according to the
    routes in the route table. The action returns an association ID, which you need in order to disassociate the
    route table later. A route table can be associated with multiple subnets.

    Args:
        name(Text): An Idem name to identify the route table resource.
        resource_id(Text, optional): AWS Route Table ID.
        route_table_id(Text): The ID of the route table.
        subnet_id(Text, optional): The ID of the subnet. A subnet ID is not returned for an implicit association.
        gateway_id(Text, optional): The ID of the internet gateway or virtual private gateway.

    Request Syntax:
        [route_table-resource-name]:
          aws.ec2.route_table_association.present:
          - route_table_id: 'string'
          - resource_id: 'string'
          - subnet_id: 'string'
          - gateway_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_route-table_association:
              aws.ec2.route_table_association.present:
              - route_table_id: vpc-02850adfa9f6fc916
              - resource_id: route_table-3485hydfe5f6tb998
              - gateway_id: vgw-0334141b528f99316
              - subnet_id: subnet-0610f7c1a12a1af6a
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx, RouteTableIds=[route_table_id]
    )
    if before and before["result"]:
        route_table_resource = before["ret"].get("RouteTables")[0]
        if resource_id:
            ret = hub.tool.aws.ec2.route_table.get_route_table_association_by_id(
                route_table_resource["Associations"], resource_id=resource_id
            )
            if ret["result"]:
                result[
                    "old_state"
                ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_association_to_present(
                    resource=ret["ret"],
                    idem_resource_name=name,
                )
                result["comment"] = (
                    f"aws.ec2.route_table_association '{name}' association already exists",
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
                return result

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "route_table_id": route_table_id,
                    "resource_id": name,
                    "subnet_id": subnet_id,
                    "gateway_id": gateway_id,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.route_table_association", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.associate_route_table(
            ctx, RouteTableId=route_table_id, SubnetId=subnet_id, GatewayId=gateway_id
        )
        if not ret["result"]:
            result["result"] = False
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.route_table_association", name=name
        )
    try:
        after = await hub.exec.boto3.client.ec2.describe_route_tables(
            ctx, RouteTableIds=[route_table_id]
        )
        resource = after["ret"].get("RouteTables")[0]
        association = hub.tool.aws.ec2.route_table.get_route_table_association_by_id(
            resource["Associations"], gateway_id=gateway_id, subnet_id=subnet_id
        )
        result[
            "new_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_association_to_present(
            resource=association["ret"],
            idem_resource_name=name,
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    route_table_id: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Disassociates a subnet or gateway from a route table. After you perform this action, the subnet no longer uses
    the routes in the route table. Instead, it uses the routes in the VPC's main route table.

    Args:
        name(Text): An Idem name to identify the route table resource.
        route_table_id(Text): The ID of the route table.
        resource_id(Text, optional): AWS Route Table ID. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.route_table_association.absent:
                - name: value
                - route_table: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.route_table_association", name=name
        )
        return result
    before = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx, RouteTableIds=[route_table_id]
    )
    if not before["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.route_table_association", name=name
        )
    else:
        if before["ret"] and before["ret"]["RouteTables"]:
            resource = before["ret"]["RouteTables"][0]
            association_ret = (
                hub.tool.aws.ec2.route_table.get_route_table_association_by_id(
                    resource["Associations"], resource_id=resource_id
                )
            )
            if association_ret["result"]:
                result[
                    "old_state"
                ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_association_to_present(
                    resource=association_ret["ret"],
                    idem_resource_name=name,
                )
                if ctx.get("test", False):
                    result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                        resource_type="aws.ec2.route_table_association", name=name
                    )
                    return result
                ret = await hub.exec.boto3.client.ec2.disassociate_route_table(
                    ctx, AssociationId=resource_id
                )
                if not ret.get("result"):
                    result["comment"] = result["comment"] + ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.ec2.route_table_association", name=name
                )
            else:
                result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                    resource_type="aws.ec2.route_table_association", name=name
                )
                return result
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your route table associations. Each subnet in your VPC must be associated with a route
    table. If a subnet is not explicitly associated with any route table, it is implicitly associated with the main
    route table. This command does not return the subnet ID for implicit associations. For more information,
    see Route tables in the Amazon Virtual Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.route_table_association
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe route_table {ret['comment']}")
        return result
    for route_table in ret["ret"]["RouteTables"]:
        if route_table["Associations"]:
            for association in route_table["Associations"]:
                if association["AssociationState"]["State"] == "associated":
                    route_table_association_id = association["RouteTableAssociationId"]
                    resource_converted = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_association_to_present(
                        resource=association,
                        idem_resource_name=route_table_association_id,
                    )
                    result[route_table_association_id] = {
                        "aws.ec2.route_table_association.present": [
                            {parameter_key: parameter_value}
                            for parameter_key, parameter_value in resource_converted.items()
                        ]
                    }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.docdb.create_db_subnet_group
hub.exec.boto3.client.docdb.delete_db_subnet_group
hub.exec.boto3.client.docdb.describe_db_subnet_groups
hub.exec.boto3.client.docdb.modify_db_subnet_group
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_subnet_group_description: str,
    subnet_ids: List,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new subnet group. subnet groups must contain at least one subnet in at least two Availability Zones in
    the Amazon Web Services Region.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): AWS DocumentDB Subnet Group Name. Defaults to None.
        db_subnet_group_description(Text): The description for the subnet group.
        subnet_ids(List): The Amazon EC2 subnet IDs for the subnet group.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}]. The tags to be assigned to the subnet group. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.docdb.db_subnet_group.present:
                - name: value
                - resource_id: value
                - db_subnet_group_description: value
                - subnet_ids: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if resource_id:
        before = await hub.exec.aws.docdb.db_subnet_group.get(
            ctx=ctx, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

    if before and before["ret"]:
        resource_arn = before["ret"]["db_subnet_group_arn"]
        old_tags = before["ret"]["tags"]

        if not old_tags:
            result["result"] = False
            result["comment"] = result["comment"] + old_tags["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        future_state = copy.deepcopy(result["old_state"])

        old_subnets = result["old_state"].get("subnet_ids", [])
        update_parameters = {
            "DBSubnetGroupName": resource_id,
            "SubnetIds": old_subnets,
        }

        # Check to see if SubNet Ids need to be updated
        if subnet_ids is not None and set(old_subnets) != set(subnet_ids):
            resource_updated = True
            update_parameters["SubnetIds"] = subnet_ids

        # Check to see if Subnet Description need to be updated
        old_description = result["old_state"].get("db_subnet_group_description")
        if (
            db_subnet_group_description is not None
            and old_description != db_subnet_group_description
        ):
            resource_updated = True
            update_parameters["DBSubnetGroupDescription"] = db_subnet_group_description

        if resource_updated:
            if ctx.get("test", False):
                future_state["subnet_ids"] = subnet_ids
                future_state[
                    "db_subnet_group_description"
                ] = db_subnet_group_description
            else:
                ret = await hub.exec.boto3.client.docdb.modify_db_subnet_group(
                    ctx, **update_parameters
                )
                result["result"] = ret["result"]
                result["comment"] = result["comment"] + ret["comment"]
                if not result["result"]:
                    return result

        # Update Tags
        if tags is not None and not hub.tool.aws.state_comparison_utils.compare_dicts(
            tags, old_tags
        ):
            resource_updated = True
            update_ret = await hub.tool.aws.docdb.tag.update_tags(
                ctx=ctx,
                resource_arn=resource_arn,
                old_tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(old_tags),
                new_tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                hub.log.debug(
                    f"Failed updating tags for aws.docdb.db_subnet_group '{name}'"
                )
                return result
            result["comment"] = result["comment"] + update_ret["comment"]
            if ctx.get("test", False) and update_ret["ret"] is not None:
                future_state["tags"] = update_ret["ret"]["tags"]

        if ctx.get("test", False):
            if resource_updated:
                result["new_state"] = future_state
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.docdb.db_subnet_group", name=name
                )
            else:
                result["comment"] = (
                    f"aws.docdb.db_subnet_group '{name}' already exists",
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        if not resource_updated:
            result["comment"] = (f"aws.docdb.db_subnet_group '{name}' already exists",)
        else:
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.docdb.db_subnet_group", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "db_subnet_group_name": name,
                    "db_subnet_group_description": db_subnet_group_description,
                    "resource_id": resource_id,
                    "subnet_ids": subnet_ids,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.docdb.db_subnet_group", name=name
            )
            return result

        ret = await hub.exec.boto3.client.docdb.create_db_subnet_group(
            ctx,
            DBSubnetGroupName=name,
            DBSubnetGroupDescription=db_subnet_group_description,
            SubnetIds=subnet_ids,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["DBSubnetGroup"]["DBSubnetGroupName"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.docdb.db_subnet_group", name=name
        )

    if (not before) or resource_updated:
        after = await hub.exec.aws.docdb.db_subnet_group.get(
            ctx=ctx, resource_id=resource_id
        )

        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a subnet group.  The specified database subnet group must not be associated with any DB instances.
    Idem automatically considers this resource being absent if this field is not specified

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, Optional): AWS DocumentDB Subnet Group Name. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.docdb.db_subnet_group.absent:
                - name: value
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.docdb.db_subnet_group", name=name
    )

    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.docdb.db_subnet_group.get(
        ctx=ctx,
        resource_id=resource_id,
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.docdb.db_subnet_group", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.docdb.db_subnet_group", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.docdb.delete_db_subnet_group(
            ctx, **{"DBSubnetGroupName": name}
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.docdb.db_subnet_group", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of DBSubnetGroup descriptions. If a DBSubnetGroupName is specified, the list will contain only
    the descriptions of the specified DBSubnetGroup.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.docdb.db_subnet_group
    """

    result = {}
    ret = await hub.exec.boto3.client.docdb.describe_db_subnet_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe db_subnet_group {ret['comment']}")
        return {}

    for db_subnet_group in ret["ret"]["DBSubnetGroups"]:
        # Including fields to match the 'present' function parameters
        resource_id = db_subnet_group.get("DBSubnetGroupName")
        resource_arn = db_subnet_group.get("DBSubnetGroupArn")
        tags = await hub.tool.aws.docdb.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not tags["result"]:
            # if error occurs when fetching tags
            hub.log.warning(
                f"Failed listing tags for aws.docdb.db_subnet_group '{resource_id}'"
                f"Describe will skip this aws.docdb.db_subnet_group and continue. "
            )
            continue
        tags = tags["ret"]
        resource_translated = (
            hub.tool.aws.docdb.db_subnet_group.convert_raw_db_subnet_group_to_present(
                db_subnet_group, idem_resource_name=resource_id, tags=tags
            )
        )

        result[resource_id] = {
            "aws.docdb.db_subnet_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



from typing import Dict
from typing import List
from typing import Tuple

# Only update this attribute when the instance is running
__update_state__ = "running"


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: Tuple[Dict[str, str]],
    new_value: Tuple[Dict[str, str]],
    comments: List[str],
) -> bool:
    """
    Bootstrap an instance if hasn't been bootstrapped already with the methods in "new_value"

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    result = True

    # There is currently no way to "undo" bootstrapping with heist
    # We only need to check for bootstrapping that hasn't been completed according to ESM
    for bootstrap_options in new_value:
        heist_manager = bootstrap_options["heist_manager"]

        if heist_manager in old_value:
            comments += [f"Already bootstrapped aws.ec2.instance with {heist_manager}"]
            continue
        elif ctx.test:
            comments += [f"Would bootstrap aws.ec2.instance with {heist_manager}"]
            continue
        else:
            bootstrap_ret = await hub.exec.aws.ec2.instance.bootstrap(
                ctx, resource.id, **bootstrap_options
            )
            result &= bootstrap_ret.result

            if isinstance(bootstrap_ret.comment, list):
                comments.extend(bootstrap_ret["comment"])
            else:
                comments.append(bootstrap_ret["comment"])

            if not result:
                return result

    return result



"""
This plugin provides an interface for interacting with the raw boto3 API using idem's acct system and CLI paradigm.

All underlying calls are done asynchronously in a Threadpool Executor
"""


def __init__(hub):
    # Provides the ctx argument to all execution modules
    # which will have profile info from the account module
    hub.exec.boto3.ACCT = ["aws"]

    # Load dynamic subs for accessing boto3 clients and resources
    hub.pop.sub.dynamic(
        sub=hub.exec.boto3,
        subname="client",
        resolver=hub.tool.boto3.resolve.client,
        context=None,
    )
    hub.pop.sub.dynamic(
        sub=hub.exec.boto3,
        subname="resource",
        resolver=hub.tool.boto3.resolve.resource,
        context=None,
    )
